"""This module provides tools for running models using yggdrasil."""
import sys
import logging
# import atexit
import os
import time
import signal
from pprint import pformat
from itertools import chain
import socket
from yggdrasil.tools import YggClass
from yggdrasil.config import ygg_cfg, cfg_environment
from yggdrasil import platform, backwards, yamlfile
from yggdrasil.drivers import create_driver


COLOR_TRACE = '\033[30;43;22m'
COLOR_NORMAL = '\033[0m'


# def setup_ygg_logging(prog, level=None):
#     r"""Set the log lovel based on environment variable 'YGG_DEBUG'. If the
#     variable is not set, the log level is set to 'NOTSET'.

#     Args:
#         prog (str): Name to prepend log messages with.
#         level (str, optional): String specifying the logging level. Defaults
#             to None and the environment variable 'YGG_DEBUG' is used.

#     """
#     if level is None:
#         level = ygg_cfg.get('debug', 'ygg', 'NOTSET')
#     logLevel = eval('logging.' + level)
#     logging.basicConfig(level=logLevel, stream=sys.stdout, format=COLOR_TRACE +
#                         prog + ': %(message)s' + COLOR_NORMAL)


class YggRunner(YggClass):
    r"""This class handles the orchestration of starting the model and
    IO drivers, monitoring their progress, and cleaning up on exit.

    Arguments:
        modelYmls (list): List of paths to yaml files specifying the models
            that should be run.
        namespace (str): Name that should be used to uniquely identify any RMQ
            exchange.
        host (str, optional): Name of the host that the models will be launched
            from. Defaults to None.
        rank (int, optional): Rank of this set of models if run in parallel.
            Defaults to 0.
        ygg_debug_level (str, optional): Level for Ygg debug messages. Defaults
            to environment variable 'YGG_DEBUG'.
        rmq_debug_level (str, optional): Level for RabbitMQ debug messages.
            Defaults to environment variable 'RMQ_DEBUG'.
        ygg_debug_prefix (str, optional): Prefix for Ygg debug messages.
            Defaults to namespace.

    Attributes:
        namespace (str): Name that should be used to uniquely identify any RMQ
            exchange.
        host (str): Name of the host that the models will be launched from.
        rank (int): Rank of this set of models if run in parallel.
        modeldrivers (dict): Model drivers associated with this run.
        inputdrivers (dict): Input drivers associated with this run.
        outputdrivers (dict): Output drivers associated with this run.
        serverdrivers (dict): The addresses associated with different server
            drivers.
        interrupt_time (float): Time of last interrupt signal.
        error_flag (bool): True if one or more models raises an error.

    ..todo:: namespace, host, and rank do not seem strictly necessary.

    """
    def __init__(self, modelYmls, namespace, host=None, rank=0,
                 ygg_debug_level=None, rmq_debug_level=None,
                 ygg_debug_prefix=None):
        super(YggRunner, self).__init__('runner')
        self.namespace = namespace
        self.host = host
        self.rank = rank
        self.modeldrivers = {}
        self.inputdrivers = {}
        self.outputdrivers = {}
        self.serverdrivers = {}
        self.interrupt_time = 0
        self._inputchannels = {}
        self._outputchannels = {}
        self._old_handlers = {}
        self.error_flag = False
        # Setup logging
        # if ygg_debug_prefix is None:
        #     ygg_debug_prefix = namespace
        # setup_ygg_logging(ygg_debug_prefix, level=ygg_debug_level)
        # Update environment based on config
        cfg_environment()
        # Parse yamls
        drivers = yamlfile.parse_yaml(modelYmls)
        self.inputdrivers = drivers['input']
        self.outputdrivers = drivers['output']
        self.modeldrivers = drivers['model']
        for x in self.outputdrivers.values():
            self._outputchannels[x['args']] = x
        for x in self.inputdrivers.values():
            self._inputchannels[x['args']] = x
        # print(pformat(self.inputdrivers), pformat(self.outputdrivers),
        #       pformat(self.modeldrivers))
        # atexit.register(self.cleanup)

    def pprint(self, *args):
        r"""Print with color."""
        s = ''.join(str(i) for i in args)
        print((COLOR_TRACE + '{}' + COLOR_NORMAL).format(s))

    def signal_handler(self, sig, frame):
        r"""Terminate all drivers on interrrupt."""
        self.debug("Interrupt with signal %d", sig)
        now = backwards.clock_time()
        elapsed = now - self.interrupt_time
        self.debug('Elapsed time since last interrupt: %d s', elapsed)
        self.interrupt_time = now
        self.pprint(' ')
        self.pprint(80 * '*')
        if elapsed < 5:
            self.pprint('* %76s *' % 'Interrupted twice within 5 seconds: shutting down')
            self.pprint(80 * '*')
            # signal.siginterrupt(signal.SIGTERM, True)
            # signal.siginterrupt(signal.SIGINT, True)
            self.debug("Terminating models and closing all channels")
            self.terminate()
            self.pprint(80 * '*')
            # self.sleep(5)
            return 1
        else:
            self.pprint('* %76s *' % 'Interrupted: Displaying channel summary')
            self.pprint('* %76s *' % 'interrupt again (within 5 seconds) to exit')
            self.pprint(80 * '*')
            self.printStatus()
            self.pprint(80 * '*')
        self.debug('%d returns', sig)

    def _swap_handler(self, signum, signal_handler):
        self._old_handlers[signum] = signal.getsignal(signum)
        signal.signal(signum, signal_handler)
        if not platform._is_win:
            signal.siginterrupt(signum, False)
        
    def set_signal_handler(self, signal_handler=None):
        r"""Set the signal handler.

        Args:
            signal_handler (function, optional): Function that should handle
                received SIGINT and SIGTERM signals. Defaults to
                self.signal_handler.

        """
        if signal_handler is None:
            signal_handler = self.signal_handler
        self._swap_handler(signal.SIGINT, signal_handler)
        if not platform._is_win:
            self._swap_handler(signal.SIGTERM, signal_handler)
        else:  # pragma: windows
            self._swap_handler(signal.SIGBREAK, signal_handler)

    def reset_signal_handler(self):
        r"""Reset signal handlers to old ones."""
        for k, v in self._old_handlers.items():
            signal.signal(k, v)

    def run(self, signal_handler=None, timer=None, t0=None):
        r"""Run all of the models and wait for them to exit.

        Args:
            signal_handler (function, optional): Function that should be used as
                a signal handler. Defaults to None and is set by
                set_signal_handler.
            timer (function, optional): Function that should be called to get
                intermediate timing statistics. Defaults to time.time if not
                provided.
            t0 (float, optional): Zero point for timing statistics. Is set
                using the provided timer if not provided.

        Returns:
            dict: Intermediate times from the run.

        """
        if timer is None:
            timer = time.time
        if t0 is None:
            t0 = timer()
        times = {}
        times['init'] = timer()
        self.loadDrivers()
        times['load drivers'] = timer()
        self.startDrivers()
        times['start drivers'] = timer()
        self.set_signal_handler(signal_handler)
        self.waitModels()
        times['run models'] = timer()
        self.reset_signal_handler()
        self.closeChannels()
        times['close channels'] = timer()
        self.cleanup()
        times['clean up'] = timer()
        tprev = t0
        key_order = ['init', 'load drivers', 'start drivers', 'run models',
                     'close channels', 'clean up']
        for k in key_order:
            self.info('%20s\t%f', k, times[k] - tprev)
            tprev = times[k]
        self.info(40 * '=')
        self.info('%20s\t%f', "Total", tprev - t0)
        return times

    @property
    def all_drivers(self):
        r"""iterator: For all drivers."""
        return chain(self.inputdrivers.values(), self.outputdrivers.values(),
                     self.modeldrivers.values())

    def io_drivers(self, model=None):
        r"""Return the input and output drivers for one or all models.

        Args:
            model (str, optional): Name of a model that I/O drivers should be
                returned for. Defaults to None and all I/O drivers are returned.

        Returns:
            iterator: Access to list of I/O drivers.

        """
        if model is None:
            out = chain(self.inputdrivers.values(), self.outputdrivers.values())
        else:
            driver = self.modeldrivers[model]
            out = chain(driver.get('input_drivers', dict()),
                        driver.get('output_drivers', dict()))
        return out

    def createDriver(self, yml):
        r"""Create a driver instance from the yaml information.

        Args:
            yml (yaml): Yaml object containing driver information.

        Returns:
            object: An instance of the specified driver.

        """
        self.debug('Creating %s, a %s', yml['name'], yml['driver'])
        curpath = os.getcwd()
        if 'ClientDriver' in yml['driver']:
            yml.setdefault('comm_address', self.serverdrivers[yml['args']])
        if 'working_dir' in yml:
            os.chdir(yml['working_dir'])
        try:
            instance = create_driver(yml=yml, namespace=self.namespace,
                                     rank=self.rank, **yml)
            yml['instance'] = instance
        finally:
            os.chdir(curpath)
        if 'ServerDriver' in yml['driver']:
            self.serverdrivers[yml['args']] = instance.comm_address
        return instance

    def createModelDriver(self, yml):
        r"""Create a model driver instance from the yaml information.

        Args:
            yml (yaml): Yaml object containing driver information.

        Returns:
            object: An instance of the specified driver.

        """
        yml['env'] = {}
        for iod in self.io_drivers(yml['name']):
            yml['env'].update(iod['instance'].env)
            iod['models'].append(yml['name'])
        drv = self.createDriver(yml)
        if 'client_of' in yml:
            for srv in yml['client_of']:
                self.modeldrivers[srv]['clients'].append(yml['name'])
        self.debug("Model %s:, env: %s",
                   yml['name'], pformat(yml['instance'].env))
        return drv

    def createInputDriver(self, yml):
        r"""Create an input driver instance from the yaml information.

        Args:
            yml (yaml): Yaml object containing driver information.

        Returns:
            object: An instance of the specified driver.

        """
        yml['models'] = []
        if yml['args'] not in self._outputchannels:
            for x in yml['icomm_kws']['comm']:
                if 'filetype' not in x:
                    raise ValueError(
                        ("Input driver %s could not locate a "
                         + "corresponding file or output channel %s") % (
                             x["name"], yml["args"]))
        drv = self.createDriver(yml)
        return drv

    def createOutputDriver(self, yml):
        r"""Create an output driver instance from the yaml information.

        Args:
            yml (yaml): Yaml object containing driver information.

        Returns:
            object: An instance of the specified driver.

        """
        yml['models'] = []
        if yml['args'] in self._inputchannels:
            yml.setdefault('comm_env', {})
            yml['comm_env'] = self._inputchannels[yml['args']]['instance'].comm_env
        if yml['args'] not in self._inputchannels:
            for x in yml['ocomm_kws']['comm']:
                if 'filetype' not in x:
                    raise ValueError(
                        ("Output driver %s could not locate a "
                         + "corresponding file or input channel %s") % (
                             x["name"], yml["args"]))
        drv = self.createDriver(yml)
        return drv
        
    def loadDrivers(self):
        r"""Load all of the necessary drivers, doing the IO drivers first
        and adding IO driver environmental variables back tot he models."""
        self.debug('')
        driver = dict(name='name')
        try:
            # Create input drivers
            self.debug("Loading input drivers")
            for driver in self.inputdrivers.values():
                self.createInputDriver(driver)
            # Create output drivers
            self.debug("Loading output drivers")
            for driver in self.outputdrivers.values():
                self.createOutputDriver(driver)
            # Create model drivers
            self.debug("Loading model drivers")
            for driver in self.modeldrivers.values():
                self.createModelDriver(driver)
        except BaseException:  # pragma: debug
            self.error("%s could not be created.", driver['name'])
            self.terminate()
            raise

    def startDrivers(self):
        r"""Start drivers, starting with the IO drivers."""
        self.info('Starting I/O drivers and models on system '
                  + '{} in namespace {} with rank {}'.format(
                      self.host, self.namespace, self.rank))
        driver = dict(name='name')
        try:
            # Start connections
            for driver in self.io_drivers():
                self.debug("Starting driver %s", driver['name'])
                d = driver['instance']
                if not d.was_started:
                    d.start()
            # Ensure connections in loop
            for driver in self.io_drivers():
                self.debug("Checking driver %s", driver['name'])
                d = driver['instance']
                d.wait_for_loop()
                assert(d.was_loop)
                assert(not d.errors)
            # Start models
            # self.sleep(1)  # on windows comms can take a while start
            for driver in self.modeldrivers.values():
                self.debug("Starting driver %s", driver['name'])
                d = driver['instance']
                for n2 in driver.get('client_of', []):
                    d2 = self.modeldrivers[n2]['instance']
                    if not d2.was_started:
                        self.debug("Starting server '%s' before client", d2.name)
                        d2.start()
                if not d.was_started:
                    d.start()
        except BaseException:  # pragma: debug
            self.error("%s did not start", driver['name'])
            self.terminate()
            raise
        self.debug('ALL DRIVERS STARTED')

    def waitModels(self):
        r"""Wait for all model drivers to finish. When a model finishes,
        join the thread and perform exits for associated IO drivers."""
        self.debug('')
        running = [d for d in self.modeldrivers.values()]
        dead = []
        while (len(running) > 0) and (not self.error_flag):
            for drv in running:
                d = drv['instance']
                if d.errors:  # pragma: debug
                    self.error('Error in model %s', drv['name'])
                    self.error_flag = True
                    break
                d.join(1)
                if not d.is_alive():
                    if not d.errors:
                        self.info("%s finished running.", drv['name'])
                        self.do_model_exits(drv)
                        self.debug("%s completed model exits.", drv['name'])
                        self.do_client_exits(drv)
                        self.debug("%s completed client exits.", drv['name'])
                        running.remove(drv)
                        self.info("%s finished exiting.", drv['name'])
                else:
                    self.info('%s still running', drv['name'])
            dead = []
            for drv in self.all_drivers:
                d = drv['instance']
                d.join(0.1)
                if not d.is_alive():
                    dead.append(drv['name'])
        for d in self.modeldrivers.values():
            if d['instance'].errors:
                self.error_flag = True
        if not self.error_flag:
            self.info('All models completed')
        else:
            self.error('One or more models generated errors.')
            self.terminate()
        self.debug('Returning')

    def do_model_exits(self, model):
        r"""Perform exits for IO drivers associated with a model.

        Args:
            model (dict): Dictionary of model parameters including any
                associated IO drivers.

        """
        for drv in self.io_drivers(model['name']):
            drv['models'].remove(model['name'])
            if not drv['instance'].is_alive():
                continue
            if (len(drv['models']) == 0):
                self.debug('on_model_exit %s', drv['name'])
                drv['instance'].on_model_exit()
    
    def do_client_exits(self, model):
        r"""Perform exits for IO drivers associated with a client model.

        Args:
            model (dict): Dictionary of model parameters including any
                associated IO drivers.

        """
        for srv_name in model.get('client_of', []):
            # Remove this client from list for server
            srv = self.modeldrivers[srv_name]
            srv['clients'].remove(model['name'])
            # Stop server if there are not any more clients
            if len(srv['clients']) == 0:
                iod = self.inputdrivers[srv_name]
                iod['instance'].on_client_exit()
                srv['instance'].stop()

    def terminate(self):
        r"""Immediately stop all drivers, beginning with IO drivers."""
        self.debug('')
        # self.closeChannels(force_stop=True)
        # self.debug('Stop models')
        for driver in self.all_drivers:
            if 'instance' in driver:
                self.debug('Stop %s', driver['name'])
                driver['instance'].terminate()
                # Terminate should ensure instance not alive
                assert(not driver['instance'].is_alive())
                # if driver['instance'].is_alive():
                #     driver['instance'].join()
        self.debug('Returning')

    def cleanup(self):
        r"""Perform cleanup operations for all drivers."""
        self.debug('')
        for driver in self.all_drivers:
            if 'instance' in driver:
                driver['instance'].cleanup()

    def printStatus(self):
        r"""Print the status of all drivers, starting with the IO drivers."""
        self.debug('')
        for driver in self.all_drivers:
            if 'instance' in driver:
                driver['instance'].printStatus()

    def closeChannels(self, force_stop=False):
        r"""Stop IO drivers and join the threads.

        Args:
            force_stop (bool, optional): If True, the terminate method is
                used to stop the drivers. Otherwise, the stop method is used.
                The stop method will try to exit gracefully while terminate
                will exit as quickly as possible. Defaults to False.

        """
        self.debug('')
        drivers = [i for i in self.io_drivers()]
        for drv in drivers:
            if 'instance' in drv:
                driver = drv['instance']
                if driver.is_alive():  # pragma: debug
                    self.debug("Stopping %s", drv['name'])
                    if force_stop or self.error_flag:
                        driver.terminate()
                    else:
                        driver.stop()
                    self.debug("Stop(%s) returns", drv['name'])
        self.debug('Channel Stops DONE')
        for drv in drivers:
            if 'instance' in drv:
                driver = drv['instance']
                assert(not driver.is_alive())
                # self.debug("Join %s", drv['name'])
                # if driver.is_alive():
                #     driver.join()
                # self.debug("Join %s done", drv['name'])
        self.debug('Returning')

        
def get_runner(models, **kwargs):
    r"""Get runner for a set of models, getting run information from the
    environment.

    Args:
        models (list): List of yaml files containing information on the models
            that should be run.
        **kwargs: Additonal keyword arguments are passed to YggRunner.

    Returns:
        YggRunner: Runner for the provided models.

    Raises:
        Exception: If config option 'namespace' in 'rmq' section not set.

    """
    # Get environment variables
    logger = logging.getLogger(__name__)
    namespace = kwargs.pop('namespace', ygg_cfg.get('rmq', 'namespace', False))
    if not namespace:  # pragma: debug
        raise Exception('rmq:namespace not set in config file')
    rank = os.environ.get('PARALLEL_SEQ', '0')
    host = socket.gethostname()
    os.environ['YGG_RANK'] = rank
    os.environ['YGG_HOST'] = host
    rank = int(rank)
    kwargs.update(rank=rank, host=host)
    # Run
    logger.debug("Running in %s with path %s namespace %s rank %d",
                 os.getcwd(), sys.path, namespace, rank)
    yggRunner = YggRunner(models, namespace, **kwargs)
    return yggRunner



import os
import glob
import copy
import six
import importlib
# import warnings
from collections import OrderedDict
from yggdrasil import backwards
from yggdrasil.doctools import docs2args


_registry = {}
_registry_defaults = {}
_registry_class2subtype = {}
_registry_complete = False
_comptype2key = {'comm': 'commtype',
                 'file': 'filetype',
                 'model': 'language',
                 'connection': 'connection_type',
                 # 'datatype': None,
                 'serializer': 'seritype'}
# 'compiler': 'toolname',
# 'linker': 'toolname',
# 'archiver': 'toolname'}
_comptype2mod = {'serializer': 'serialize',
                 'comm': 'communication',
                 'file': 'communication',
                 'model': 'drivers',
                 'connection': 'drivers'}
# 'datatype': ['metaschema', 'datatypes'],
# 'compiler': 'drivers',
# 'linker': 'drivers',
# 'archiver': 'drivers'}


def init_registry():
    r"""Initialize the registries and schema."""
    global _registry
    global _registry_complete
    if not _registry_complete:
        comp_list = []
        mod_list = []
        for k, v in _comptype2mod.items():
            if v not in mod_list:
                comp_list.append(k)
                mod_list.append(v)
        for k in comp_list:
            import_all_components(k)
        _registry_complete = True
    return _registry


# This dosn't work as desried because classes that have already been imported
# will not call registration on second import
# def clear_registry():
#     r"""Reset registries."""
#     global _registry
#     global _registry_defaults
#     global _registry_class2subtype
#     global _registry_complete
#     _registry = {}
#     _registry_defaults = {}
#     _registry_class2subtype = {}
#     _registry_complete = False

    
def suspend_registry():
    r"""Suspend the registry by storing the global registries in a dictionary."""
    global _registry
    global _registry_defaults
    global _registry_class2subtype
    global _registry_complete
    out = {'_registry': _registry, '_registry_defaults': _registry_defaults,
           '_registry_class2subtype': _registry_class2subtype,
           '_registry_complete': _registry_complete}
    _registry = {}
    _registry_defaults = {}
    _registry_class2subtype = {}
    _registry_complete = False
    return out


def restore_registry(reg_dict):
    r"""Restore the registry to values in the provided dictionary."""
    global _registry
    global _registry_defaults
    global _registry_class2subtype
    global _registry_complete
    _registry = reg_dict['_registry']
    _registry_defaults = reg_dict['_registry_defaults']
    _registry_class2subtype = reg_dict['_registry_class2subtype']
    _registry_complete = reg_dict['_registry_complete']


def import_all_components(comptype):
    r"""Dynamically import all component classes for a component type.

    Args:
        comptype (str): Component type.

    """
    # Get module and directory
    mod = copy.deepcopy(_comptype2mod[comptype])
    moddir = copy.deepcopy(_comptype2mod[comptype])
    # The next three lines will be required if there are ever any components
    # nested in multiple directories (e.g. metaschema/datatypes)
    # if isinstance(mod, list):
    #     mod = '.'.join(mod)
    #     moddir = os.path.join(*moddir)
    moddir = os.path.join(os.path.dirname(__file__), moddir)
    modbase = importlib.import_module('yggdrasil.%s' % mod)
    non_comp = [os.path.splitext(x)[0] for x in
                getattr(modbase, '_non_component_modules', [])]
    # Import all files
    for x in glob.glob(os.path.join(moddir, '*.py')):
        xbase = os.path.splitext(os.path.basename(x))[0]
        if (not xbase.startswith('__')) and (xbase not in non_comp):
            importlib.import_module('yggdrasil.%s.%s'
                                    % (mod, xbase))


def import_component(comptype, subtype=None, without_schema=False):
    r"""Dynamically import a component by name.

    Args:
        comptype (str): Component type.
        subtype (str, optional): Component subtype. If subtype is not one of the
            registered subtypes for the specified comptype, subtype is treated
            as the name of class. Defaults to None if not provided and the
            default subtype defined in the schema for the specified component
            will be used.
        without_schema (bool, optional): If True, the schema is not used to
            import the component and subtype must be the name of a component
            class. Defaults to False. subtype must be provided if without_schema
            is True.

    Returns:
        class: Component class.

    Raises:
        ValueError: If subtype is not provided, but without_schema is True.
        ValueError: If comptype is not a registered component type.
        ValueError: If subtype is not a registered subtype or the name of a
            registered subtype class for the specified comptype.
            

    """
    # Get module
    mod = _comptype2mod[comptype]
    # if isinstance(mod, list):
    #     mod = '.'.join(mod)
    # Set direct import shortcuts for unregistered classes
    if (comptype == 'comm') and (subtype is None):
        subtype = 'DefaultComm'
    if (((comptype in ['comm', 'file']) and (subtype is not None)
         and ((subtype == 'CommBase') or subtype.endswith('Comm')))):
        without_schema = True
    # Set default based on registry to avoid schema if possible
    if (subtype is None) and (comptype in _registry_defaults):
        subtype = _registry_defaults.get(comptype, None)
    # Check registered components to prevent importing multiple times
    if subtype in _registry.get(comptype, {}):
        out_cls = _registry[comptype][subtype]
    elif subtype in _registry_class2subtype.get(comptype, {}):
        out_cls = _registry[comptype][_registry_class2subtype[comptype][subtype]]
    else:
        # Get class name
        if without_schema:
            if subtype is None:  # pragma: debug
                raise ValueError("subtype must be provided if without_schema is True.")
            class_name = subtype
        else:
            from yggdrasil.schema import get_schema
            s = get_schema().get(comptype, None)
            if s is None:  # pragma: debug
                raise ValueError("Unrecognized component type: %s" % comptype)
            if subtype is None:  # pragma: no cover
                # This will only be called if the test is run before the component
                # module is imported
                subtype = s.default_subtype
            if subtype in s.class2subtype:
                class_name = subtype
            else:
                class_name = s.subtype2class.get(subtype, None)
                if class_name is None:
                    # Attempt file since they are subclass of comm
                    if (comptype == 'comm'):
                        try:
                            return import_component('file', subtype)
                        except ValueError:
                            pass
                    raise ValueError("Unrecognized %s subtype: %s"
                                     % (comptype, subtype))
        try:
            out_mod = importlib.import_module('yggdrasil.%s.%s' % (mod, class_name))
        except ImportError:  # pragma: debug
            import_all_components(comptype)
            return import_component(comptype, subtype=subtype,
                                    without_schema=without_schema)
        out_cls = getattr(out_mod, class_name)
    # Check for an aliased class
    if hasattr(out_cls, '_get_alias'):
        out_cls = out_cls._get_alias()
    return out_cls


def create_component(comptype, subtype=None, **kwargs):
    r"""Dynamically create an instance of a component with the specified options
    as outlined in the component schema. This function requires loading the
    component schemas and so should not be used at the module or class level to
    prevent circular dependencies.

    Args:
        comptype (str): Component type.
        subtype (str, optional): Component subtype. If subtype is not one of the
            registered subtypes for the specified comptype, subtype is treated
            as the name of the class. Defaults to None if not provided and the
            default subtype defined in the schema for the specified component
            will be used. If the subtype is specified by the component subtype
            key in the remaining kwargs, that subtype will be used instead.
        **kwargs: Additional keyword arguments are treated as options for the
            component as outlined in the component schema.

    Returns:
        ComponentBase: Instance of the specified component type/subtype and
            options.
    
    Raises:
        ValueError: If comptype is not a registered component type.

    """
    from yggdrasil.schema import get_schema
    s = get_schema().get(comptype, None)
    if s is None:  # pragma: debug
        raise ValueError("Unrecognized component type: %s" % comptype)
    if s.subtype_key in kwargs:
        subtype = kwargs[s.subtype_key]
    cls = import_component(comptype, subtype=subtype)
    return cls(**kwargs)


def inherit_schema(orig, new_properties=None, new_required=None,
                   remove_keys=[], **kwargs):
    r"""Create an inherited schema, adding new value to accepted ones for
    dependencies.
    
    Args:
        orig (dict): Schema that will be inherited.
        new_properties (dict, optional): Dictionary of new properties to add.
            Defaults to None and is ignored.
        new_requried (list, optional): Properties that should be required by the
            new class. Defaults to None and is ignored.
        remove_keys (list, optional): Keys that should be removed form orig
            before adding the new keys. Defaults to empty list.
        **kwargs: Additional keyword arguments will be added to the schema
            with dependency on the provided key/value pair.

    Returns:
        tuple(dict, list): New schema properties and a list of requried.

    """
    remove_keys = copy.deepcopy(remove_keys)
    # Get set of original properties
    assert(issubclass(orig, ComponentBase))
    out_prp = copy.deepcopy(orig._schema_properties)
    if orig._schema_excluded_from_inherit is not None:
        remove_keys += orig._schema_excluded_from_inherit
    out_req = copy.deepcopy(orig._schema_required)
    # Don't add duplicates
    if new_properties == out_prp:
        new_properties = None
    if new_required == out_req:
        new_required = None
    # Remove keys
    for k in remove_keys:
        if k in out_prp:
            out_prp.pop(k)
        if k in out_req:
            out_req.remove(k)
    # Add new values and keyword arguments
    if new_properties is not None:
        out_prp.update(new_properties)
    if new_required is not None:
        for k in new_required:
            if k not in out_req:
                out_req.append(k)
    out_prp.update(kwargs)
    return out_prp, out_req


class ComponentMeta(type):
    r"""Meta class for registering schema components."""
    
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        # Return early for error classes which should be unregistered duplicates
        # of the base class
        if getattr(cls, '_is_error_class', False):
            return cls
        # Determine subtype
        subtype = None
        if cls._schema_subtype_key is not None:
            subtype = getattr(cls, cls._schema_subtype_key,
                              getattr(cls, '_' + cls._schema_subtype_key, None))
        # Inherit new schema properties
        if cls._schema_inherit and (cls.__name__ != 'ComponentBase'):
            assert(isinstance(cls._schema_inherit, bool))
            inherit_from = None
            for x in bases:
                if hasattr(x, '_schema_properties'):
                    inherit_from = x
                    break
            if inherit_from is None:  # pragma: debug
                raise RuntimeError(("Class %s dosn't have a component "
                                    "parent class.") % cls)
            # Dont inherit if the base is ComponentBase (empty schema)
            if inherit_from.__name__ != ComponentBase:
                cls._schema_properties, cls._schema_required = inherit_schema(
                    inherit_from,
                    new_properties=cls._schema_properties,
                    new_required=cls._schema_required,
                    remove_keys=cls._schema_excluded_from_inherit)
        # Do non-registration things
        if not (name.endswith('Base') or (subtype is None) or cls._dont_register):
            # Perform class specific actions in preparation for registration
            cls.before_registration(cls)
            # assert(cls.name is not None)
            # Add parameter descriptions from docs
            for x in cls.__mro__[::-1]:
                args_dict = docs2args(x.__doc__)
                for k, v in cls._schema_properties.items():
                    if k in args_dict:
                        v.setdefault('description', args_dict[k]['description'])
            # Register
            global _registry
            global _registry_defaults
            global _registry_class2subtype
            yaml_typ = cls._schema_type
            default_subtype = cls._schema_properties.get(
                cls._schema_subtype_key, {}).get('default',
                                                 cls._schema_subtype_default)
            if yaml_typ not in _registry:
                _registry[yaml_typ] = OrderedDict()
                _registry_defaults[yaml_typ] = default_subtype
                _registry_class2subtype[yaml_typ] = {}
            elif default_subtype is not None:
                assert(_registry_defaults[yaml_typ] == default_subtype)
            if cls.__name__ not in _registry[yaml_typ]:
                _registry[yaml_typ][cls.__name__] = cls
                _registry_class2subtype[yaml_typ][subtype] = cls.__name__
            if not (os.environ.get('YGG_RUNNING_YGGSCHEMA', 'None').lower()
                    in ['true', '1']):
                cls.after_registration(cls)
                cls.finalize_registration(cls)
        return cls
    
    # def __getattribute__(cls, key):
    #     r"""If the class is an alias for another class and has been initialized,
    #     call getattr on the aliased class."""
    #     if key not in ['__dict__', '_get_alias']:
    #         if hasattr(cls, '_get_alias') and (key not in cls.__dict__):
    #             return getattr(cls._get_alias(), key)
    #     return super(ComponentMeta, cls).__getattribute__(key)


@six.add_metaclass(ComponentMeta)
class ComponentBase(object):
    r"""Base class for schema components.

    Args:
        skip_component_schema_normalization (bool, optional): If True, the
            schema will not be used to normalize/validate input keyword
            arguments (e.g. in case they were already parsed). Defaults to
            False.
        **kwargs: Keyword arguments are added to the class as attributes
            according to the class attributes _schema_properties and
            _schema_excluded_from_class. Keyword arguments not added to the
            class as attributes are assigned to the extra_kwargs dictionary.

    Attributes:
        extra_kwargs (dict): Keyword arguments that were not parsed.

    Class Attributes:
        _schema_type (str): Name of the component type the class represents.
        _schema_subtype_key (str): Attribute that should be used to identify the
            subtype associated with each class.
        _schema_subtype_description (str): Description for the subtype represented
            by the class that should be used in documentation tables.
        _schema_required (list): Keys from _schema_properties that are required
            to produce a valid component.
        _schema_properties (dict): Schemas describing keyword arguments that
            can be supplied to the class constructor and used to specify
            component behavior in YAML/JSON files. At initialization, these
            keywords are added to the class instance as attributes of the
            same name unless they are in _schema_excluded_from_class. Unless
            _schema_inherit is False, these properties will be added in addition
            to the schema properties defined by the class base.
        _schema_excluded_from_class (list): Keywords in _schema_properties that
            should not be added to the class as attributes during initialization.
        _schema_excluded_from_inherit (list): Keywords in _schema_properties that
            should not be inherited either from the base class or by child
            classes.
        _schema_inherit (bool, ComponentBase): If False, the base schema will
            not be inherited. If a Component subclass, the schema from that
            class will be inherited instead of the base class. Defaults to True.
        _dont_register (bool): If True, the component class will be be registered
            and the before_registration class method will not be called. Defaults
            to False.
        

    """

    _schema_type = None
    _schema_subtype_key = None
    _schema_subtype_description = None
    _schema_subtype_default = None
    _schema_required = []
    _schema_properties = {}
    _schema_excluded_from_class = []
    _schema_excluded_from_inherit = []
    _schema_excluded_from_class_validation = []
    _schema_inherit = True
    _dont_register = False
    
    def __init__(self, skip_component_schema_normalization=None, **kwargs):
        if skip_component_schema_normalization is None:
            skip_component_schema_normalization = (
                os.environ.get('YGG_SKIP_COMPONENT_VALIDATION', 'None').lower()
                in ['true', '1'])
        comptype = self._schema_type
        if (comptype is None) and (not self._schema_properties):
            self.extra_kwargs = kwargs
            return
        subtype = None
        if self._schema_subtype_key is not None:
            subtype = getattr(self, self._schema_subtype_key,
                              getattr(self, '_%s' % self._schema_subtype_key, None))
        # Fall back to some simple parsing/normalization to save time on
        # full jsonschema normalization
        for k, v in self._schema_properties.items():
            if k in self._schema_excluded_from_class:
                continue
            default = v.get('default', None)
            if (k == self._schema_subtype_key) and (subtype is not None):
                default = subtype
            if default is not None:
                kwargs.setdefault(k, copy.deepcopy(default))
            if v.get('type', None) == 'array':
                if isinstance(kwargs.get(k, None), backwards.string_types):
                    kwargs[k] = kwargs[k].split()
        # Parse keyword arguments using schema
        if (comptype is not None) and (subtype is not None):
            from yggdrasil.schema import get_schema
            s = get_schema().get_component_schema(comptype, subtype, relaxed=True)
            props = list(s['properties'].keys())
            if not skip_component_schema_normalization:
                from yggdrasil import metaschema
                kwargs.setdefault(self._schema_subtype_key, subtype)
                # Remove properties that shouldn't ve validated in class
                for k in self._schema_excluded_from_class_validation:
                    if k in s['properties']:
                        del s['properties'][k]
                # Validate and normalize
                metaschema.validate_instance(kwargs, s, normalize=False)
                # TODO: Normalization performance needs improvement
                # import pprint
                # print('before')
                # pprint.pprint(kwargs_comp)
                # kwargs_comp = metaschema.validate_instance(kwargs_comp, s,
                #                                            normalize=True)
                # kwargs.update(kwargs_comp)
                # print('normalized')
                # pprint.pprint(kwargs_comp)
        else:
            props = self._schema_properties.keys()
        # Set attributes based on properties
        for k in props:
            if k in self._schema_excluded_from_class:
                continue
            v = kwargs.pop(k, None)
            if getattr(self, k, None) is None:
                setattr(self, k, v)
            # elif (getattr(self, k) != v) and (v is not None):
            #     warnings.warn(("The schema property '%s' is provided as a "
            #                    "keyword with a value of %s, but the class "
            #                    "already has an attribute of the same name "
            #                    "with the value %s.")
            #                   % (k, v, getattr(self, k)))
        self.extra_kwargs = kwargs

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration. These actions will still be performed if the environment
        variable YGG_RUNNING_YGGSCHEMA is set."""
        pass

    @staticmethod
    def after_registration(cls):
        r"""Operations that should be preformed to modify class attributes after
        registration. These actions will not be performed if the environment
        variable YGG_RUNNING_YGGSCHEMA is set."""
        pass

    @staticmethod
    def finalize_registration(cls):
        r"""Final operations to perform after a class has been fully initialized.
        These actions will not be performed if the environment variable
        YGG_RUNNING_YGGSCHEMA is set."""
        pass



import types
from yggdrasil.metaschema.datatypes.ClassMetaschemaType import (
    ClassMetaschemaType)


class FunctionMetaschemaType(ClassMetaschemaType):
    r"""Type for evaluating functions."""

    name = 'function'
    description = 'Type for callable Python functions.'
    python_types = (types.BuiltinFunctionType, types.FunctionType,
                    types.BuiltinMethodType, types.MethodType)



r"""Routines for interfacing with framework from Python."""
# These import from the symbolic link for backwards compat
from yggdrasil.interface import YggInterface
from yggdrasil.interface.YggInterface import (
    YGG_MSG_EOF, YggInput, YggOutput)


__all__ = ['YggInterface', 'YggInput', 'YggOutput', 'YGG_MSG_EOF']



import os
import copy
import glob
from yggdrasil import components
from yggdrasil.drivers.CompiledModelDriver import CompiledModelDriver


class BuildModelDriver(CompiledModelDriver):
    r"""Class for running build file compiled drivers.

    Args:
        name (str): Driver name.
        args (str, list): Executable that should be created (make target) and
            any arguments for the executable.
        target (str, optional): Make target that should be built to create the
            model executable. Defaults to None.
        target_language (str, optional): Language that the target is written in.
            Defaults to None and will be set based on the source files provided.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        buildfile (str): Path to file containing build instructions.
        builddir (str): Path to directory where build products will be saved.
        sourcedir (str): Path to directory where source files are located.
        compiledir (str): Path to directory where compilation call should be
            made from. Defaults to working_dir.
        target (str): Name of executable that should be created and called.
        target_language (str): Language that the target is written in.
        target_language_driver (ModelDriver): Language driver for the target
            language.

    Class Attributes:
        built_where_called (bool): If True, it is assumed that compilation
            output will be saved in the same directory from which the
            compilation command is issued.

    Raises:
        RuntimeError: If neither the IPC or ZMQ C libraries are available.

    """

    base_languages = ['c', 'c++']
    built_where_called = False
    sourcedir_as_sourcefile = False

    def __init__(self, *args, **kwargs):
        self.target_language_driver = None
        super(BuildModelDriver, self).__init__(*args, **kwargs)

    def parse_arguments(self, args, **kwargs):
        r"""Sort arguments based on their syntax to determine if an argument
        is a source file, compilation flag, or runtime option/flag that should
        be passed to the model executable.

        Args:
            args (list): List of arguments provided.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        """
        # Set builddir before passing to parent class so that builddir is used
        # to normalize the model file path rather than the working directory
        # which may be different.
        default_attr = [('target_language_driver', None),
                        ('buildfile', None),
                        ('builddir', None),
                        ('sourcedir', None),
                        ('compile_working_dir', None),
                        ('builddir_base', '.'),
                        ('buildfile_base', None)]
        for k, v in default_attr:
            if not hasattr(self, k):
                setattr(self, k, v)
        # Directory that compilation should be called from
        if ((isinstance(self.compile_working_dir, str)
             and not os.path.isabs(self.compile_working_dir))):
            self.compile_working_dir = os.path.realpath(
                os.path.join(self.working_dir, self.compile_working_dir))
        # Source directory
        if self.sourcedir is None:
            self.sourcedir = os.path.dirname(args[0])
        if not os.path.isabs(self.sourcedir):
            self.sourcedir = os.path.realpath(os.path.join(self.working_dir,
                                                           self.sourcedir))
        # Build file
        if self.buildfile is None:
            self.buildfile = self.buildfile_base
        if not os.path.isabs(self.buildfile):
            for x in set([self.working_dir, self.sourcedir,
                          self.builddir, self.compile_working_dir]):
                if x is not None:
                    y = os.path.normpath(os.path.join(x, self.buildfile))
                    if os.path.isfile(y):
                        self.buildfile = y
                        break
        # Compilation directory
        if self.compile_working_dir is None:
            self.compile_working_dir = os.path.dirname(self.buildfile)
        if not os.path.isabs(self.compile_working_dir):
            self.compile_working_dir = os.path.realpath(
                os.path.join(self.working_dir, self.compile_working_dir))
        # Build directory
        if self.builddir is None:
            if self.built_where_called:
                self.builddir = self.compile_working_dir
            else:
                self.builddir = os.path.join(self.sourcedir, self.builddir_base)
        if not os.path.isabs(self.builddir):
            self.builddir = os.path.realpath(os.path.join(self.working_dir,
                                                          self.builddir))
        kwargs.setdefault('default_model_dir', self.builddir)
        if self.sourcedir_as_sourcefile:
            self.source_files = [self.sourcedir]
        super(BuildModelDriver, self).parse_arguments(args, **kwargs)

    def set_target_language(self):
        r"""Set the language of the target being compiled (usually the same
        as the language associated with this driver.

        Returns:
            str: Name of language.

        """
        source_dir = self.sourcedir
        if self.target_language is None:
            # Get file form cmakelists files
            try_list = sorted(list(glob.glob(os.path.join(source_dir, '*'))))
            early_exit = False
            if self.model_src is not None:
                try_list = [self.model_src, try_list]
                early_exit = True
            languages = copy.deepcopy(self.get_tool_instance('compiler').languages)
            languages.remove(self.language)
            try:
                self.target_language = self.get_language_for_source(
                    try_list, early_exit=early_exit, languages=languages)
            except ValueError:
                pass
            # Try to compile C as C++
            # if self.target_language == 'c':
            #     self.target_language = 'c++'
        if (((self.target_language_driver is None)
             and (self.target_language is not None))):
            self.target_language_driver = components.import_component(
                'model', self.target_language)
        return self.target_language
        
    @classmethod
    def is_source_file(cls, fname):
        r"""Determine if the provided file name points to a source files for
        the associated programming language by checking the extension.

        Args:
            fname (str): Path to file.

        Returns:
            bool: True if the provided file is a source file, False otherwise.

        """
        compiler = cls.get_tool('compiler')
        for lang in compiler.languages:
            if lang == cls.language:
                continue
            drv = components.import_component('model', lang)
            if drv.is_source_file(fname):
                return True
        return False
        
    def compile_model(self, **kwargs):
        r"""Compile model executable(s).

        Args:
            **kwargs: Keyword arguments are passed on to the parent class's
                method.

        """
        kwargs['working_dir'] = self.compile_working_dir
        return super(BuildModelDriver, self).compile_model(**kwargs)
        
    def cleanup(self):
        r"""Remove compiled executable."""
        if (self.model_file is not None) and os.path.isfile(self.model_file):
            self.compile_model(target='clean')
        super(BuildModelDriver, self).cleanup()



from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)


class TestItemsMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for ItemsMetaschemaProperty class."""
    
    _mod = 'JSONArrayMetaschemaProperties'
    _cls = 'ItemsMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestItemsMetaschemaProperty, self).__init__(*args, **kwargs)
        nele = 3
        valid_value = [int(i) for i in range(nele)]
        valid_sing = {'type': 'int'}
        valid_mult = [{'type': 'int'} for i in range(nele)]
        invalid_sing = {'type': 'float'}
        invalid_mult = [{'type': 'float'} for i in range(nele)]
        self._valid = [(valid_value, valid_sing),
                       (valid_value, valid_mult),
                       ([int(i) for i in range(nele - 1)], valid_sing)]
        self._invalid = [([float(i) for i in range(nele)], valid_sing),
                         ([float(i) for i in range(nele)], valid_mult)]
        # ([int(i) for i in range(nele - 1)], valid_mult)]
        self._valid_compare = [(valid_sing, valid_sing),
                               (valid_sing, valid_mult),
                               (valid_mult, valid_sing),
                               (valid_mult, valid_mult)]
        self._invalid_compare = [(valid_sing, invalid_sing),
                                 (valid_sing, invalid_mult),
                                 (valid_mult, invalid_sing),
                                 (valid_mult, invalid_mult),
                                 (1, 1),
                                 (valid_mult, valid_mult[:-1])]



import os
import glob
import importlib
from collections import OrderedDict


_metaschema_properties = OrderedDict()


def register_metaschema_property(prop_class):
    r"""Register a schema property.

    Args:
        prop_class (class): Class to be registered.

    Raises:
        ValueError: If a property class has already been registered under the
            same name.
        ValueError: If the base validator already has a validator function for
            the property and the new property class has a schema defined.
        ValueError: If the base validator already has a validator function for
            the property and the validate method on the new property class is
            not disabled.
        ValueError: If the property class does not have an entry in the existing
            metaschema.

    """
    from yggdrasil.metaschema import _metaschema, _base_validator
    global _metaschema_properties
    prop_name = prop_class.name
    if prop_name in _metaschema_properties:
        raise ValueError("Property '%s' already registered." % prop_name)
    if prop_name in _base_validator.VALIDATORS:
        if (prop_class.schema is not None):
            raise ValueError("Replacement property '%s' modifies the default schema."
                             % prop_name)
        if (((prop_class._validate not in [None, False])
             or ('validate' in prop_class.__dict__))):
            raise ValueError("Replacement property '%s' modifies the default validator."
                             % prop_name)
        prop_class._validate = False
    # prop_class.types = []  # To ensure base class not modified by all
    # prop_class.python_types = []
    # Check metaschema if it exists
    if _metaschema is not None:
        if prop_name not in _metaschema['properties']:
            raise ValueError("Property '%s' not in pre-loaded metaschema." % prop_name)
    _metaschema_properties[prop_name] = prop_class
    return prop_class


class MetaschemaPropertyMeta(type):
    r"""Meta class for registering properties."""

    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        if not (name.endswith('Base') or (cls.name in ['base'])
                or cls._dont_register):
            cls = register_metaschema_property(cls)
        return cls
        

def get_registered_properties():
    r"""Return a dictionary of registered properties.

    Returns:
        dict: Registered property/class pairs.

    """
    return _metaschema_properties


def get_metaschema_property(property_name, skip_generic=False):
    r"""Get the property class associated with a metaschema property.

    Args:
        property_name (str): Name of property to get class for.
        skip_generic (bool, optional): If True and the property dosn't have a
            class, None is returned. Defaults to False.

    Returns:
        MetaschemaProperty: Associated property class.

    """
    from yggdrasil.metaschema.properties import MetaschemaProperty
    if property_name in _metaschema_properties:
        return _metaschema_properties[property_name]
    else:
        if skip_generic:
            return None
        else:
            return MetaschemaProperty.MetaschemaProperty


def import_all_properties():
    r"""Import all types to ensure they are registered."""
    for x in glob.glob(os.path.join(os.path.dirname(__file__), '*.py')):
        mod = os.path.basename(x)[:-3]
        if not mod.startswith('__'):
            importlib.import_module('yggdrasil.metaschema.properties.%s' % mod)



from yggdrasil import languages, tools, platform
from yggdrasil.tests import assert_raises


def test_get_language_dir():
    r"""Test the get_language_dir method."""
    assert_raises(ValueError, languages.get_language_dir, 'invalid')
    test_lang = tools.get_supported_lang()
    test_skip = ['make', 'cmake', 'executable']
    for lang in test_lang:
        if lang in test_skip:
            continue
        languages.get_language_dir(lang)


def test_get_language_ext():
    r"""Test the get_language_ext method."""
    test_lang = tools.get_supported_lang()
    for lang in test_lang:
        if (lang == 'executable') and (not platform._is_win):
            assert_raises(ValueError, languages.get_language_ext, lang)
            languages.get_language_ext(lang, default='')
        else:
            languages.get_language_ext(lang)



import os
import unittest
import signal
import uuid
from yggdrasil import runner, tools, platform
from yggdrasil.tests import YggTestBase, assert_raises
# from yggdrasil.tests import yamls as sc_yamls
from yggdrasil.examples import yamls as ex_yamls


def test_get_runner():
    r"""Use get_runner to start a run."""
    namespace = "test_get_runner_%s" % str(uuid.uuid4)
    cr = runner.get_runner([ex_yamls['hello']['python']],
                           namespace=namespace)
    cr.run()
    cr.sleep()


# def test_runner_error():
#     r"""Start a runner for a model with an error."""
#     cr = runner.get_runner([sc_yamls['error']])
#     cr.run()


# Spawning fake Ctrl-C works locally for windows, but causes hang on appveyor
@unittest.skipIf(platform._is_win, "Signal processing not sorted on windows")
def test_Arunner_interrupt():
    r"""Start a runner then stop it with a keyboard interrupt."""
    cr = runner.get_runner([ex_yamls['hello']['python']])
    if platform._is_win:  # pragma: debug
        cr.debug_log()
    cr.loadDrivers()
    cr.startDrivers()
    cr.set_signal_handler()
    tools.kill(os.getpid(), signal.SIGINT)
    tools.kill(os.getpid(), signal.SIGINT)
    cr.reset_signal_handler()
    cr.waitModels()
    cr.closeChannels()
    cr.cleanup()
    if platform._is_win:  # pragma: debug
        cr.reset_log()


def test_runner_terminate():
    r"""Start a runner, then stop it early."""
    cr = runner.get_runner([ex_yamls['hello']['python']])
    cr.loadDrivers()
    cr.startDrivers()
    cr.printStatus()
    cr.terminate()


def test_runner_error():
    r"""Test error on missing yaml."""
    assert_raises(IOError, runner.YggRunner,
                  ['fake_yaml.yml'], 'test_ygg_run')
    

class TestYggRunner(YggTestBase):
    r"""Tests of the YggRunner class."""
    def setup(self, *args, **kwargs):
        super(TestYggRunner, self).setup(*args, **kwargs)
        self.runner = runner.YggRunner([ex_yamls['hello']['python']],
                                       'test_ygg_run')

    def test_createIODriver(self):
        r"""Test createInputDriver and createOutputDriver."""
        yml = {'name': 'fake_IODriver',
               'args': 'fake_channel',
               'driver': 'InputDriver',
               'working_dir': os.getcwd(),
               'icomm_kws': {'comm': [{'name': 'fake_IODriver'}]},
               'ocomm_kws': {'comm': [{'name': 'fake_IODriver'}]},
               'kwargs': {}}
        assert_raises(Exception, self.runner.createInputDriver, yml)
        yml['driver'] = 'OutputDriver'
        assert_raises(Exception, self.runner.createOutputDriver, yml)



import os
import unittest
import logging
from yggdrasil.tests import scripts, assert_raises
import yggdrasil.drivers.tests.test_InterpretedModelDriver as parent
from yggdrasil import runner
from yggdrasil.drivers import MatlabModelDriver
from yggdrasil.examples import yamls as ex_yamls


logger = logging.getLogger(__name__)
_session_fname = os.path.join(os.getcwd(), 'nt_screen_session.txt')


def test_is_matlab_running():
    r"""Test if there is Matlab engine running."""
    MatlabModelDriver.is_matlab_running()
    MatlabModelDriver.kill_all()
    assert(not MatlabModelDriver.is_matlab_running())


@unittest.skipIf(MatlabModelDriver._matlab_engine_installed,
                 "Matlab installed.")
def test_matlab_not_installed():  # pragma: no matlab
    r"""Assert that errors are raised when Matlab is not installed."""
    assert_raises(RuntimeError, MatlabModelDriver.start_matlab_engine)
    assert_raises(RuntimeError, MatlabModelDriver.stop_matlab_engine,
                  None, None, None, None)
    assert_raises(RuntimeError, MatlabModelDriver.MatlabProcess, None, None)


@unittest.skipIf(not MatlabModelDriver.MatlabModelDriver.is_installed(),
                 "Matlab not installed.")
def test_matlab_runner():  # pragma: matlab
    r"""Use get_runner to start a Matlab run."""
    cr = runner.get_runner([ex_yamls['hello']['matlab']])
    cr.run()


@unittest.skipIf(not MatlabModelDriver.MatlabModelDriver.is_installed(),
                 "Matlab not installed.")
def test_matlab_exit():  # pragma: matlab
    r"""Test error when model contains 'exit' call."""
    MatlabModelDriver.MatlabModelDriver('error', [scripts['matlab_error']])
    # Re-enable if it becomes necessary to raise an error instead of just a warning
    # assert_raises(RuntimeError, MatlabModelDriver.MatlabModelDriver, 'error',
    #                  [scripts['matlab_error']])


@unittest.skipIf(not MatlabModelDriver.MatlabModelDriver.is_installed(),
                 "Matlab not installed.")
def test_locate_matlabroot():  # pragma: matlab
    r"""Test locate_matlabroot."""
    MatlabModelDriver.locate_matlabroot()


class TestMatlabModelParam(parent.TestInterpretedModelParam):  # pragma: matlab
    r"""Test parameters for MatlabModelDriver."""

    driver = "MatlabModelDriver"
    
    def __init__(self, *args, **kwargs):
        super(TestMatlabModelParam, self).__init__(*args, **kwargs)
        self.args = [self.src[0], "test", 1]
        self.attr_list += ['started_matlab', 'mlengine']

            
class TestMatlabModelDriverNoInit(TestMatlabModelParam,  # pragma: matlab
                                  parent.TestInterpretedModelDriverNoInit):
    r"""Test runner for MatlabModelDriver without instance."""
    pass


class TestMatlabModelDriverNoStart(TestMatlabModelParam,  # pragma: matlab
                                   parent.TestInterpretedModelDriverNoStart):
    r"""Test runner for MatlabModelDriver without starting the driver."""
    pass


class TestMatlabModelDriver(TestMatlabModelParam,
                            parent.TestInterpretedModelDriver):  # pragma: matlab
    r"""Test runner for MatlabModelDriver."""
    
    @unittest.skipIf(not MatlabModelDriver._matlab_engine_installed,
                     "Matlab engine not installed.")
    def test_a(self):
        r"""Dummy test to start matlab."""
        if self.instance.screen_session is None:  # pragma: debug
            logger.info("Matlab was not started by this test. Close any "
                        + "existing Matlab sessions to test creation/removal.")
        else:
            with open(_session_fname, 'w') as f:
                f.write(self.instance.screen_session)
            self.instance.screen_session = None
            self.instance.started_matlab = False

    @unittest.skipIf(not MatlabModelDriver._matlab_engine_installed,
                     "Matlab engine not installed.")
    def test_z(self):
        r"""Dummy test to stop matlab."""
        if os.path.isfile(_session_fname):
            with open(_session_fname, 'r') as f:
                session = f.read()
            os.remove(_session_fname)
            self.instance.screen_session = session
            self.instance.started_matlab = True
        else:  # pragma: debug
            logger.info("Skipping removal of Matlab session as the test did "
                        + "not create it.")



import uuid
import threading
from yggdrasil.communication import CommBase


class AsyncTryAgain(Exception):
    r"""Exception raised when comm open, but send should be attempted again."""
    pass


class AsyncComm(CommBase.CommBase):
    r"""Class for handling asynchronous I/O.

    Args:
        name (str): The name of the message queue.
        dont_backlog (bool, optional): If True, the backlog will not be started
            and all messages will be sent/received directly to/from the comm.
            Defaults to False.
        **kwargs: Additional keyword arguments are passed to CommBase.
        
    Attributes:
        dont_backlog (bool): If True, the backlog will not be started and all
            messages will be sent/received directly to/from the comm.
        backlog_send_ready (threading.Event): Event set when there is a
            message in the send backlog.
        backlog_recv_ready (threading.Event): Event set when there is a
            message in the recv backlog.
        
    """
    def __init__(self, name, dont_backlog=False, **kwargs):
        # TODO: Fix the cleanup of Python threads in languages that call the
        # Python API underneath
        self.dont_backlog = (dont_backlog or kwargs.get('is_interface', False))
        self._backlog_recv = []
        self._backlog_send = []
        self._backlog_thread = None
        self.backlog_send_ready = threading.Event()
        self.backlog_recv_ready = threading.Event()
        self.backlog_open = False
        self._used_direct = False
        super(AsyncComm, self).__init__(name, **kwargs)

    def get_status_message(self, nindent=0):
        r"""Return lines composing a status message.
        
        Args:
            nindent (int, optional): Number of tabs that should be used to
                indent each line. Defaults to 0.
                
        Returns:
            tuple(list, prefix): Lines composing the status message and the
                prefix string used for the last message.

        """
        lines, prefix = super(AsyncComm, self).get_status_message(nindent=nindent)
        lines += ['%s%-15s: %s' % (prefix, 'open (backlog)', self.is_open_backlog),
                  '%s%-15s: %s' % (prefix, 'open (direct)', self.is_open_direct),
                  '%s%-15s: %s' % (prefix, 'nsent (backlog)', self.n_msg_backlog_send),
                  '%s%-15s: %s' % (prefix, 'nrecv (backlog)', self.n_msg_backlog_recv),
                  '%s%-15s: %s' % (prefix, 'nsent (direct)', self.n_msg_direct_send),
                  '%s%-15s: %s' % (prefix, 'nrecv (direct)', self.n_msg_direct_recv)]
        if len(self._work_comms) > 0:
            lines.append('%sWork comms:' % prefix)
            for v in self._work_comms.values():
                lines += v.get_status_message(nindent=nindent + 1)[0]
        return lines, prefix
        
    @property
    def backlog_thread(self):
        r"""tools.YggThread: Thread that will handle sinding or receiving
        backlogged messages."""
        if self._backlog_thread is None:
            if self.direction == 'send':
                self._backlog_thread = CommBase.CommThreadLoop(
                    self, target=self.run_backlog_send, suffix='SendBacklog')
            else:
                self._backlog_thread = CommBase.CommThreadLoop(
                    self, target=self.run_backlog_recv, suffix='RecvBacklog')
        return self._backlog_thread

    def open(self):
        r"""Open the connection by connecting to the queue."""
        super(AsyncComm, self).open()
        self._open_direct()
        if self.is_open_direct:
            self._open_backlog()

    def _open_direct(self):
        r"""Open the comm directly."""
        pass

    def _open_backlog(self):
        r"""Open the backlog."""
        if not self.is_open_backlog:
            self.backlog_open = True
            if not self.dont_backlog:
                self.backlog_thread.start()

    def _close_direct(self, linger=False):
        r"""Close the comm directly."""
        pass

    def _close_backlog(self, wait=False):
        r"""Close the backlog thread."""
        self.debug('')
        self.backlog_open = False
        self.backlog_thread.set_break_flag()
        self.backlog_send_ready.set()
        self.backlog_recv_ready.set()
        if wait and not self.dont_backlog:
            self.backlog_thread.wait(key=str(uuid.uuid4()))

    def _close(self, linger=False):
        r"""Close the connection.

        Args:
            linger (bool, optional): If True, drain messages before closing the
                comm. Defaults to False.

        """
        self.debug('')
        self._close_backlog(wait=True)
        self._close_direct()
        super(AsyncComm, self)._close(linger=linger)

    def stop_backlog(self):
        r"""Stop the asynchronous backlog, turning this into a direct comm."""
        self._close_backlog(wait=True)
        self.dont_backlog = True

    @property
    def is_open(self):
        r"""bool: True if the backlog is open."""
        return self.is_open_direct or self.is_open_backlog

    @property
    def is_open_direct(self):
        r"""bool: True if the direct comm is not None."""
        return False

    @property
    def is_open_backlog(self):
        r"""bool: True if the backlog thread is running."""
        return self.backlog_open

    @property
    def n_msg_direct_recv(self):
        r"""int: Number of messages currently being routed in recv."""
        return 0

    @property
    def n_msg_direct_send(self):
        r"""int: Number of messages currently being routed in send."""
        return 0

    @property
    def n_msg_direct(self):
        r"""int: Number of messages currently being routed."""
        if self.direction == 'send':
            return self.n_msg_direct_send
        else:
            return self.n_msg_direct_recv

    @property
    def n_msg_backlog_recv(self):
        r"""int: Number of messages in the receive backlog."""
        if self.is_open_backlog:
            return len(self.backlog_recv)
        return 0

    @property
    def n_msg_backlog_send(self):
        r"""int: Number of messages in the send backlog."""
        if self.is_open_backlog:
            return len(self.backlog_send)
        return 0

    @property
    def n_msg_backlog(self):
        r"""int: Number of messages in the backlog."""
        if self.direction == 'recv':
            return self.n_msg_backlog_recv
        else:
            return self.n_msg_backlog_send

    @property
    def n_msg_recv(self):
        r"""int: Number of messages in the receive backlog."""
        if self.direction == 'recv':
            if self.dont_backlog:
                return self.n_msg_direct_recv
            else:
                return self.n_msg_backlog_recv
        else:
            return self.n_msg_direct_recv

    @property
    def n_msg_send(self):
        r"""int: Number of messages in the send backlog."""
        if self.direction == 'send':
            if self.dont_backlog:
                return self.n_msg_direct_send
            else:
                return self.n_msg_backlog_send
        else:
            return self.n_msg_direct_send

    @property
    def n_msg_recv_drain(self):
        r"""int: Number of messages in the receive backlog and direct comm."""
        return self.n_msg_direct_recv + self.n_msg_backlog_recv

    @property
    def n_msg_send_drain(self):
        r"""int: Number of messages in the send backlog and direct comm."""
        return self.n_msg_direct_send + self.n_msg_backlog_send

    @property
    def is_confirmed_send(self):
        r"""bool: True if all sent messages have been confirmed."""
        for v in self._work_comms.values():
            if (v.direction == 'send') and not v.is_confirmed_send:  # pragma: debug
                return False
        return (self.n_msg_direct_send == 0)

    @property
    def is_confirmed_recv(self):
        r"""bool: True if all received messages have been confirmed."""
        for v in self._work_comms.values():
            if (v.direction == 'recv') and not v.is_confirmed_recv:  # pragma: debug
                return False
        return (self.n_msg_direct_recv == 0)

    @property
    def backlog_recv(self):
        r"""list: Messages that have been received."""
        with self.backlog_thread.lock:
            return self._backlog_recv

    @property
    def backlog_send(self):
        r"""list: Messages that should be sent."""
        with self.backlog_thread.lock:
            return self._backlog_send

    def add_backlog_recv(self, msg):
        r"""Add a message to the backlog of received messages.

        Args:
            msg (str): Received message that should be backlogged.

        """
        with self.backlog_thread.lock:
            self.debug("Added %d bytes to recv backlog.", len(msg))
            self._backlog_recv.append(msg)
            self.backlog_recv_ready.set()

    def add_backlog_send(self, msg, **kwargs):
        r"""Add a message to the backlog of messages to be sent.

        Args:
            msg (str): Message that should be backlogged for sending.
            **kwargs: Additional keyword arguments are added along with
                the message.

        """
        with self.backlog_thread.lock:
            self.debug("Added %d bytes to send backlog.", len(msg))
            self._backlog_send.append((msg, kwargs))
            self.backlog_send_ready.set()

    def pop_backlog_recv(self):
        r"""Pop a message from the front of the recv backlog.

        Returns:
            str: First backlogged recv message.

        """
        with self.backlog_thread.lock:
            msg = self._backlog_recv.pop(0)
            self.debug("Popped %d bytes from recv backlog.", len(msg))
            if len(self._backlog_recv) == 0:
                self.backlog_recv_ready.clear()
        return msg

    def pop_backlog_send(self):
        r"""Pop a message from the front of the send backlog.

        Returns:
            tuple (str, dict): First backlogged send message and
                keyword arguments.

        """
        with self.backlog_thread.lock:
            msg, kwargs = self._backlog_send.pop(0)
            self.debug("Popped %d bytes from send backlog.", len(msg))
            if len(self._backlog_send) == 0:
                self.backlog_send_ready.clear()
        return msg, kwargs

    def run_backlog_send(self):
        r"""Continue trying to send buffered messages."""
        if not self.is_open_backlog:  # pragma: debug
            self.debug("Backlog closed")
            self._close_backlog()
            return
        if not self.send_backlog():  # pragma: debug
            self.debug("Stopping because send_backlog failed")
            self._close_backlog()
            return
        self.periodic_debug('run_backlog_send', period=1000)(
            "Sleeping (is_confirmed_send=%s)",
            str(self.is_confirmed_send))
        self.sleep()

    def run_backlog_recv(self):
        r"""Continue buffering received messages."""
        if self.backlog_thread.main_terminated:  # pragma: debug
            self.debug("Main thread terminated")
            self._close_backlog()
            self.close()
        if not self.is_open_backlog:  # pragma: debug
            self.debug("Backlog closed")
            self._close_backlog()
            return
        if not self.recv_backlog():
            # Stop the thread, but don't close the backlog
            self.debug("Stopping backlog recv thread")
            self.backlog_thread.set_break_flag()
            return
        self.periodic_debug('run_backlog_recv', period=1000)(
            "Sleeping (is_confirmed_recv=%s)",
            str(self.is_confirmed_recv))
        self.sleep()

    def send_backlog(self):
        r"""Send a message from the send backlog to the queue."""
        if len(self.backlog_send) == 0:
            self.confirm_send()
            return True
        try:
            imsg, ikwargs = self.backlog_send[0]
            if not self._used_direct:
                self.suppress_special_debug = True
            flag = self._send_direct(imsg, **ikwargs)
            self.suppress_special_debug = False
            if flag:
                self.debug("Sent %d bytes to %s", len(imsg), self.address)
                self.pop_backlog_send()
                self._used_direct = True
        except AsyncTryAgain:  # pragma: debug
            flag = True
        except BaseException:  # pragma: debug
            self.exception('Error sending backlogged message')
            flag = False
        self.confirm_send()
        return flag

    def recv_backlog(self):
        r"""Check for any messages in the queue and add them to the recv
        backlog."""
        if not self.is_open_direct:
            self.debug("Direct comm closed.")
            flag = False
        elif self.n_msg_direct_recv == 0:
            self.verbose_debug("No messages waiting.")
            flag = True
        else:
            try:
                if not self._used_direct:
                    self.suppress_special_debug = True
                flag, data = self._recv_direct()
                self.suppress_special_debug = False
                if flag and data:
                    self.debug("Recv %d bytes from %s", len(data), self.address)
                    self.add_backlog_recv(data)
                    self._used_direct = True
            except BaseException:  # pragma: debug
                self.exception('Error receiving into backlog.')
                flag = False
        self.confirm_recv()
        return flag

    def _send_direct(self, payload):  # pragma: debug
        r"""Send a message to the comm directly.

        Args:
            payload (str): Message to send.

        Returns:
            bool: Success or failure of sending the message.

        """
        return False

    def _recv_direct(self):  # pragma: debug
        r"""Receive a message from the comm directly.

        Returns:
            tuple (bool, str): The success or failure of receiving a message
                and the message received.

        """
        return (False, self.empty_bytes_msg)

    def _send(self, payload, no_backlog=False, no_confirm=False, **kwargs):
        r"""Send a message to the backlog.

        Args:
            payload (str): Message to send.
            no_backlog (bool, optional): If False, any messages that can't be
                sent because the queue is full will be added to a list of
                messages to be sent once the queue is no longer full. If True,
                messages are not backlogged and an error will be raised if the
                queue is full. Defaults to False.
            no_confirm (bool, optional): If False and no_backlog is True, then
                this will block until the sent message is confirmed. If True,
                this will return without confirmation. Defaults to False.

        Returns:
            bool: Success or failure of sending the message.

        """
        if not self.is_open_direct:  # pragma: debug
            return False
        if self.dont_backlog:
            no_backlog = True
        if self.direction == 'recv':
            self.debug("Receive comm sending %d bytes direct.", len(payload))
            no_backlog = True
        if no_backlog or not self.backlog_send_ready.is_set():
            try:
                if not self._used_direct:
                    self.suppress_special_debug = True
                out = self._send_direct(payload, **kwargs)
                self.suppress_special_debug = False
                if out:
                    self._used_direct = True
                if no_backlog:
                    if out and (self.direction == 'send'):
                        out = self.wait_for_confirm(active_confirm=True,
                                                    timeout=False,
                                                    noblock=no_confirm,
                                                    direction='send')
                    return out
                elif out:
                    return out
            except AsyncTryAgain:
                if no_backlog:  # pragma: debug
                    raise
        self.add_backlog_send(payload, **kwargs)
        self.debug('%d bytes backlogged', len(payload))
        return True

    def _recv(self, timeout=None, no_backlog=False, no_confirm=False):
        r"""Receive a message from the backlog.

        Args:
            timeout (float, optional): Time in seconds to wait for a message.
                Defaults to self.recv_timeout.
            no_backlog (bool, optional): If False and there are messages in the
                receive backlog, they will be returned first. Otherwise the
                queue is checked for a message. Defaults to False.
            no_confirm (bool, optional): If False and no_backlog is True, then
                this will block until the sent message is confirmed. If True,
                this will return without confirmation. Defaults to False.

        Returns:
            tuple (bool, str): The success or failure of receiving a message
                and the message received.

        """
        if timeout is None:
            timeout = self.recv_timeout
        if self.dont_backlog:
            no_backlog = True
        if self.direction == 'send':
            self.debug("Send comm receiving direct.")
            no_backlog = True
        # If no backlog, receive from queue
        if no_backlog:
            T = self.start_timeout(timeout, key_suffix='_recv:direct')
            while ((not T.is_out) and (self.n_msg_direct_recv == 0)
                   and self.is_open_direct):
                self.sleep()
            self.stop_timeout(key_suffix='_recv:direct')
            if not self.is_open_direct:  # pragma: debug
                self.debug("Comm closed")
                return (False, self.empty_bytes_msg)
            if self.n_msg_direct_recv == 0:  # pragma: debug
                self.verbose_debug("No messages waiting.")
                return (True, self.empty_bytes_msg)
            if not self._used_direct:
                self.suppress_special_debug = True
            out = self._recv_direct()
            self.suppress_special_debug = False
            if out:
                self._used_direct = True
            if out and (self.direction == 'recv'):
                self.wait_for_confirm(active_confirm=True,
                                      timeout=False,
                                      noblock=no_confirm,
                                      direction='recv')
            return out
        # Sleep until there is a message
        T = self.start_timeout(timeout, key_suffix='_recv:backlog')
        while (not T.is_out) and (not self.backlog_recv_ready.is_set()):
            self.backlog_recv_ready.wait(self.sleeptime)
        self.stop_timeout(key_suffix='_recv:backlog')
        # Return False if the queue is closed
        if (not self.is_open_backlog):  # pragma: debug
            self.debug("Backlog closed")
            return (False, self.empty_bytes_msg)
        # Return True, '' if there are no messages
        if not self.backlog_recv_ready.is_set():
            self.verbose_debug("No messages waiting.")
            return (True, self.empty_bytes_msg)
        # Return backlogged message
        self.debug('Returning backlogged received message')
        return (True, self.pop_backlog_recv())

    def purge(self):
        r"""Purge all messages from the comm."""
        super(AsyncComm, self).purge()
        with self.backlog_thread.lock:
            if self.direction == 'recv':
                while self.n_msg_direct > 0:  # pragma: debug
                    self._recv_direct()
            self.backlog_recv_ready.clear()
            self.backlog_send_ready.clear()
            self._backlog_recv = []
            self._backlog_send = []



from yggdrasil.drivers.ConnectionDriver import ConnectionDriver


class OutputDriver(ConnectionDriver):
    r"""Driver for sending output to another model's comm via a local comm.

    Args:
        name (str): The name of the local message comm that the driver should
            connect to.
        args (str): The name of the other message comme that the driver should
            connect to.
        **kwargs: Additional keyword arguments are passed to parent class.

    """

    _connection_type = 'output'
    _direction = 'output'
    _schema_subtype_description = ('Connection between a model '
                                   'and one or more comms/files.')

    def __init__(self, name, args, **kwargs):
        kwargs.setdefault('ocomm_kws', {})
        kwargs['ocomm_kws']['name'] = args
        if kwargs['ocomm_kws'].get('comm', self._ocomm_type) == 'RMQComm':
            kwargs['ocomm_kws']['queue'] = args
        super(OutputDriver, self).__init__(name, **kwargs)



import time


def calc_shoot_mass(r_s, dt, S_t, R_t, R_tp1):
    r"""Calculate the shoot mass.

    Args:
        r_s (float): Relative shoot growth rate.
        dt (float): The time step.
        S_t (float): Previous shoot mass.
        R_t (float): Previous root mass.
        R_tp1 (float): Root mass at the next timestep.

    Returns:
        float: Shoot mass at the next timestep.

    """
    time.sleep(0.1)  # To simulate a longer calculation
    return (S_t * r_s * dt) + S_t - (R_tp1 - R_t)



import numpy as np
import copy
import warnings
from yggdrasil import units, backwards
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType
from yggdrasil.metaschema.datatypes.FixedMetaschemaType import (
    create_fixed_type_class)
from yggdrasil.metaschema.properties import (
    get_metaschema_property, ScalarMetaschemaProperties)


class ScalarMetaschemaType(MetaschemaType):
    r"""Type associated with a scalar.

    Developer Notes:
        |yggdrasil| defines scalars as an umbrella type encompassing int, uint,
        float, bytes, and unicode.

    """

    name = 'scalar'
    description = 'A scalar value with or without units.'
    properties = ['subtype', 'precision', 'units']
    definition_properties = ['subtype']
    metadata_properties = ['subtype', 'precision', 'units']
    extract_properties = ['subtype', 'precision', 'units']
    python_types = ScalarMetaschemaProperties._all_python_scalars

    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        if super(ScalarMetaschemaType, cls).validate(units.get_data(obj),
                                                     raise_errors=raise_errors):
            dtype = ScalarMetaschemaProperties.data2dtype(obj)
            if cls.is_fixed and ('subtype' in cls.fixed_properties):
                type_list = [
                    ScalarMetaschemaProperties._valid_types[
                        cls.fixed_properties['subtype']]]
            else:
                type_list = ScalarMetaschemaProperties._valid_numpy_types
            for k in type_list:
                if dtype.name.startswith(k):
                    return True
            else:
                if raise_errors:
                    raise ValueError(("dtype %s dosn't corresponding with any "
                                      + "of the accepted types: %s") %
                                     (str(dtype), str(type_list)))
        return False
        
    # @classmethod
    # def coerce_type(cls, obj, typedef=None, **kwargs):
    #     r"""Coerce objects of specific types to match the data type.

    #     Args:
    #         obj (object): Object to be coerced.
    #         typedef (dict, optional): Type defintion that object should be
    #             coerced to. Defaults to None.
    #         **kwargs: Additional keyword arguments are metadata entries that may
    #             aid in coercing the type.

    #     Returns:
    #         object: Coerced object.

    #     """
    #     if ((cls.is_fixed and ('subtype' in cls.fixed_properties)
    #          and (cls.fixed_properties['subtype'] in ['bytes', 'unicode']))):
    #         obj = cls.normalize(obj)
    #     return obj
        
    @classmethod
    def normalize(cls, obj):
        r"""Normalize an object, if possible, to conform to this type.

        Args:
            obj (object): Object to normalize.

        Returns:
            object: Normalized object.

        """
        if cls.is_fixed and ('subtype' in cls.fixed_properties):
            if (cls.fixed_properties['subtype'] == 'bytes'):
                if isinstance(obj, backwards.string_types):
                    obj = backwards.as_bytes(obj)
                else:
                    obj = backwards.as_bytes(str(obj))
            elif (cls.fixed_properties['subtype'] == 'unicode'):
                if isinstance(obj, backwards.string_types):
                    obj = backwards.as_unicode(obj)
                else:
                    obj = backwards.as_unicode(str(obj))
            else:
                dtype = ScalarMetaschemaProperties._python_scalars[
                    cls.fixed_properties['subtype']][0]
                try:
                    obj = dtype(obj)
                except (TypeError, ValueError):
                    pass
        return obj

    @classmethod
    def encode_data(cls, obj, typedef):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        arr = cls.to_array(obj)
        bytes = arr.tobytes()
        out = backwards.base64_encode(bytes).decode('ascii')
        return out

    @classmethod
    def encode_data_readable(cls, obj, typedef):
        r"""Encode an object's data in a readable format that may not be
        decoded in exactly the same way.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        arr = cls.to_array(obj)
        if isinstance(typedef, dict):
            subtype = typedef.get('subtype', typedef.get('type', None))
        else:
            subtype_cls = get_metaschema_property('subtype')
            subtype = subtype_cls.encode(obj)
        if (cls.name in ['1darray', 'ndarray']):
            return arr.tolist()
        assert(arr.ndim > 0)
        if subtype in ['int', 'uint']:
            return int(arr[0])
        elif subtype in ['float']:
            return float(arr[0])
        elif subtype in ['complex']:
            return str(complex(arr[0]))
        elif subtype in ['bytes', 'unicode']:
            return str(backwards.as_str(arr[0]))
        else:  # pragma: debug
            warnings.warn(("No method for handling readable serialization of "
                           + "subtype '%s', falling back to default.") % subtype)
            return super(ScalarMetaschemaType, cls).encode_data_readable(obj, typedef)

    @classmethod
    def decode_data(cls, obj, typedef):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        bytes = backwards.base64_decode(obj.encode('ascii'))
        dtype = ScalarMetaschemaProperties.definition2dtype(typedef)
        arr = np.frombuffer(bytes, dtype=dtype)
        # arr = np.fromstring(bytes, dtype=dtype)
        if 'shape' in typedef:
            arr = arr.reshape(typedef['shape'])
        out = cls.from_array(arr, unit_str=typedef.get('units', None), dtype=dtype)
        # Cast numpy type to native python type if they are equivalent
        out = cls.as_python_type(out, typedef)
        return out

    @classmethod
    def transform_type(cls, obj, typedef=None):
        r"""Transform an object based on type info.

        Args:
            obj (object): Object to transform.
            typedef (dict, optional): Type definition that should be used to
                transform the object. Defaults to None.

        Returns:
            object: Transformed object.

        """
        if typedef is None:
            return obj
        typedef0 = cls.encode_type(obj)
        typedef1 = copy.deepcopy(typedef0)
        typedef1.update(**typedef)
        dtype = ScalarMetaschemaProperties.definition2dtype(typedef1)
        arr = cls.to_array(obj).astype(dtype, casting='same_kind')
        out = cls.from_array(arr, unit_str=typedef0.get('units', None), dtype=dtype)
        out = cls.as_python_type(out, typedef)
        return units.convert_to(out, typedef1.get('units', None))

    @classmethod
    def to_array(cls, obj):
        r"""Get np.array representation of the data.

        Args:
            obj (object): Object to get array for.

        Returns:
            np.ndarray: Array representation of object.

        """
        obj_nounits = units.get_data(obj)
        if isinstance(obj_nounits, np.ndarray):
            arr = obj_nounits
        else:
            dtype = ScalarMetaschemaProperties.data2dtype(obj_nounits)
            arr = np.array([obj_nounits], dtype=dtype)
        return arr
        
    @classmethod
    def from_array(cls, arr, unit_str=None, dtype=None):
        r"""Get object representation of the data.

        Args:
            arr (np.ndarray): Numpy array.
            unit_str (str, optional): Units that should be added to returned
                object.
            dtype (np.dtype, optional): Numpy data type that should be maintained
                as a base class when adding units. Defaults to None and is
                determined from the object.

        Returns:
            object: Object representation of the data in the input array.

        """
        if (cls.name not in ['1darray', 'ndarray']) and (arr.ndim > 0):
            out = arr[0]
        else:
            out = arr
        if unit_str is not None:
            if dtype is None:
                dtype = ScalarMetaschemaProperties.data2dtype(out)
            out = units.add_units(out, unit_str, dtype=dtype)
        return out

    @classmethod
    def get_extract_properties(cls, metadata):
        r"""Get the list of properties that should be kept when extracting a
        typedef from message metadata.

        Args:
            metadata (dict): Metadata that typedef is being extracted from.

        Returns:
            list: Keywords that should be kept in the typedef.

        """
        out = super(ScalarMetaschemaType, cls).get_extract_properties(metadata)
        dtype = metadata.get('subtype', metadata['type'])
        if (((dtype in ScalarMetaschemaProperties._flexible_types)
             and (metadata['type'] not in ['1darray', 'ndarray'])
             and (not metadata.get('fixed_precision', False)))):
            out.remove('precision')
        if units.is_null_unit(metadata.get('units', '')):
            out.remove('units')
        return out

    @classmethod
    def as_python_type(cls, obj, typedef):
        r"""Convert a possible numpy type into a native Python type if possible.

        Args:
            obj (object): Object to convert.
            typedef (dict): Type definition for the object.

        Returns:
            object: Native Python version of input object if conversion possible.

        """
        if ((isinstance(typedef, dict)
             and (typedef.get('type', '1darray') not in ['1darray', 'ndarray']))):
            stype = typedef.get('subtype', typedef.get('type', None))
            py_type = ScalarMetaschemaProperties._python_scalars[stype][0]
            if np.dtype(py_type) == type(obj):
                obj = py_type(obj)
        return obj


# Dynamically create explicity scalar classes for shorthand
for t in ScalarMetaschemaProperties._valid_types.keys():
    short_doc = 'A %s value with or without units.' % t
    long_doc = ('%s\n\n'
                '    Developer Notes:\n'
                '        Precision X is preserved.\n\n') % short_doc
    kwargs = {'target_globals': globals(),
              '__doc__': long_doc,
              'python_types': ScalarMetaschemaProperties._python_scalars[t]}
    create_fixed_type_class(t, short_doc, ScalarMetaschemaType,
                            {'subtype': t}, **kwargs)



import os
import sys
import uuid
import tempfile
import subprocess
import logging
import argparse
PY_MAJOR_VERSION = sys.version_info[0]
logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)


name_in_pragmas = 'R'
lang_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
desc_file = os.path.join(lang_dir, 'R', 'DESCRIPTION')


def update_argparser(parser=None):
    r"""Update argument parser with language specific arguments.

    Args:
        parser (argparse.ArgumentParser, optional): Existing argument parser
            that should be updated. Default to None and a new argument parser
            will be created.

    Returns:
        argparse.ArgumentParser: Argument parser with language specific arguments.

    """
    if parser is None:
        parser = argparse.ArgumentParser("Run R installation script.")
    parser.add_argument('--sudoR', action='store_true', dest='sudo',
                        help='Run R installation steps with sudo.')
    parser.add_argument('--skip-r-requirements', '--skip_r_requirements',
                        action='store_true',
                        help='Don\'t install dependencies.')
    parser.add_argument('--update-r-requirements', '--update_r_requirements',
                        action='store_true',
                        help='Update the requirements.')
    return parser


def write_makevars(fname=None):
    r"""Write a makevars file with CC, CFLAGS, etc. values set based on the
    environment variables of the same name.

    Args:
        fname (str, optional): Full path to file where the files should be
            saved. Defaults to os.path.join("~", ".R", "Makevars").

    Returns:
        tuple(str, str): Full path to file where Makevars was written and the
            file indicated by the previous value of the R_MAKEVARS_USER
            environment variable. None is returned for either/both if a file is
            not written and/or the environment variable wasn't set.

    """
    if fname is None:
        fname = os.path.expanduser(os.path.join("~", ".R", "Makevars_temp"))
        if sys.platform in ['win32', 'cygwin']:
            fname += '.win'
    if os.path.isfile(fname):
        logger.info("Makevars file already exists: %s" % fname)
        return None, None
    lines = []
    for x in ['CC', 'CFLAGS', 'CXX', 'CXXFLAGS']:
        env = os.environ.get(x, '')
        if not env:
            continue
        lines.append('%s=%s' % (x, env))
    if lines:
        logger.info("Writing Makevars to %s" % fname)
        if not os.path.isdir(os.path.dirname(fname)):
            os.mkdir(os.path.dirname(fname))
        with open(fname, 'w') as fd:
            fd.write('\n'.join(lines))
        old_makevars = os.environ.get('R_MAKEVARS_USER', None)
        os.environ['R_MAKEVARS_USER'] = fname
        return fname, old_makevars
    else:
        logger.info("Nothing to be written to the Makevars file")
    return None, None


def restore_makevars(makevars, old_makevars):
    r"""Restore original makevars environemnt variable and remove temporary file.

    Args:
       makevars (str): Full path to the file where the temporary Makevars file
           was written. If None, nothing is done.
       old_makevars (str): Full path to the file where the old Makevars file
           is as defined by the environment variable R_MAKEVARS_USER.

    """
    if makevars is None:
        return
    if os.path.isfile(makevars):
        os.remove(makevars)
    if old_makevars:
        os.environ['R_MAKEVARS_USER'] = old_makevars
    else:
        if 'R_MAKEVARS_USER' in os.environ:
            del os.environ['R_MAKEVARS_USER']


def install_packages(package_list, update=False, repos=None, **kwargs):
    r"""Install R packages from CRAN.

    Args:
        package_list (str, list): One or more R packages that should be
            installed.
        update (bool, optional): If True, existing packages will be removed
            and then re-installed. If False, nothing will be done for existing
            packages. Defaults to False.
        repos (str, optional): Mirror where packages should be installed from.
            Defaults to 'http://cloud.r-project.org'.
        **kwargs: Additional keyword arguments are passed to call_R.

    Returns:
        bool: True if call was successful, False otherwise.

    """
    if not isinstance(package_list, list):
        package_list = [package_list]
    req_list = 'c(%s)' % ', '.join(['\"%s\"' % x for x in package_list])
    if repos is None:
        repos = 'http://cloud.r-project.org'
    if update:
        # R_cmd = ['install.packages(%s, repos="%s")' % (req_list, repos)]
        R_cmd = ['req <- %s' % req_list,
                 'for (x in req) {',
                 '  if (is.element(x, installed.packages()[,1])) {',
                 '    remove.packages(x)',
                 '  }',
                 '  install.packages(x, dep=TRUE, repos="%s")' % repos,
                 '}']
    else:
        R_cmd = ['req <- %s' % req_list,
                 'for (x in req) {',
                 '  if (!is.element(x, installed.packages()[,1])) {',
                 '    install.packages(x, dep=TRUE, repos="%s")' % repos,
                 '  } else {',
                 '    print(sprintf("%s already installed.", x))',
                 '  }',
                 '}']
    if not call_R(R_cmd, **kwargs):
        logger.error("Error installing dependencies: %s" % ', '.join(package_list))
        return False
    logger.info("Installed dependencies: %s" % ', '.join(package_list))
    return True

            
def call_R(R_cmd, **kwargs):
    r"""Call R commands, checking output.

    Args:
        R_cmd (list): List of R commands to run.
        **kwargs: Additional keyword arguments are passed to make_call.

    Returns:
        bool: True if the call was successful, False otherwise.

    """
    R_script = os.path.join(tempfile.gettempdir(),
                            'wrapper_%s.R' % (str(uuid.uuid4()).replace('-', '_')))
    with open(R_script, 'w') as fd:
        fd.write('\n'.join(R_cmd))
        logger.info('Running:\n    ' + '\n    '.join(R_cmd))
    try:
        out = make_call(['Rscript', R_script], **kwargs)
    finally:
        os.remove(R_script)
    return out


def make_call(R_cmd, with_sudo=False, **kwargs):
    r"""Call command, checking output.

    Args:
        R_cmd (list): List of command line executable to call and any arguments
           that should be passed to it.
        with_sudo (bool, optional): If True, the R installation script will be
            called with sudo. Defaults to False. Only valid for unix style
            operating systems.
        **kwargs: Additional keyword arguments are passed to subprocess.check_output.

    Returns:
        bool: True if the call was successful, False otherwise.

    """
    if with_sudo and (sys.platform not in ['win32', 'cygwin']):
        R_cmd.insert(0, 'sudo')
    try:
        logger.info("Calling %s on %s (with_sudo=%s)"
                    % (' '.join(R_cmd), sys.platform, with_sudo))
        R_proc = subprocess.check_output(R_cmd, **kwargs)
        if PY_MAJOR_VERSION == 3:
            R_proc = R_proc.decode("utf-8")
        logger.info("Output:\n%s" % R_proc)
        out = True
    except BaseException as e:
        logger.error('Error installing R interface:\n%s' % e)
        out = False
    return out


def requirements_from_description(fname=None):
    r"""Read R requirements from the Imports & Depends sections of the package
    description.

    Args:
        fname (str, optional): Full path to the description file. Defaults to
            desc_file.

    Returns:
        list: List of R requirements from the Imports & Depends sections.

    """
    out = []
    in_section = False
    if fname is None:
        fname = desc_file
    assert(os.path.isfile(fname))
    with open(fname, 'r') as fd:
        for x in fd.readlines():
            if x.startswith(('Imports:', 'Depends:')):
                in_section = True
            elif in_section:
                if x.startswith('  '):
                    out.append(x.strip().strip(','))
                else:
                    in_section = False
    out = list(set(out))
    return out


def install(args=None, with_sudo=None, skip_requirements=None,
            update_requirements=None):
    r"""Attempt to install the R interface.

    Args:
        args (argparse.Namespace, optional): Arguments parsed from the
            command line. Default to None and is created from sys.argv.
        with_sudo (bool, optional): If True, the R installation script will be
            called with sudo. Defaults to None and will be set based on args
            and environment variable YGG_USE_SUDO_FOR_R. Only valid for unix
            style operating systems.
        skip_requirements (bool, optional): If True, the requirements will not
            be installed. Defaults to None and is set based on if the flag
            '--skip-r-requirements' is in args.
        update_requirements (bool, optional): If True, the requirements will be
            updated. Defaults to False. Setting this to True, sets
            skip_requirements to False.

    Returns:
        bool: True if install succeded, False otherwise.

    """
    # Parse input
    if args is None:
        args = update_argparser().parse_args()
    if with_sudo is None:
        with_sudo = ((os.environ.get('YGG_USE_SUDO_FOR_R', '0') == '1')
                     or args.sudo or ('sudo' in sys.argv))
        # or args.sudoR)
    if skip_requirements is None:
        skip_requirements = args.skip_r_requirements
    if update_requirements is None:
        update_requirements = args.update_r_requirements
    if update_requirements:
        skip_requirements = False
    # Set platform dependent things
    if sys.platform in ['win32', 'cygwin']:
        R_exe = 'R.exe'
    else:
        R_exe = 'R'
    kwargs = {'cwd': lang_dir, 'with_sudo': with_sudo}
    # Write Makevars for conda installation
    makevars = None
    old_makevars = None
    if os.environ.get('CONDA_PREFIX', ''):
        makevars, old_makevars = write_makevars()
    try:
        # Install requirements
        if not skip_requirements:
            requirements = requirements_from_description()
            if not install_packages(requirements, update=update_requirements, **kwargs):
                restore_makevars(makevars, old_makevars)
                return False
            logger.info("Installed dependencies.")
        # Check to see if yggdrasil installed
        # Build packages
        build_cmd = [R_exe, 'CMD', 'build', 'R']
        if not make_call(build_cmd, **kwargs):
            logger.error("Error building R interface.")
            restore_makevars(makevars, old_makevars)
            return False
        logger.info("Built R interface.")
        # Install package
        package_name = 'yggdrasil_0.1.tar.gz'
        R_call = ("install.packages(\"%s\", "
                  "repos=NULL, type=\"source\")") % package_name
        if not call_R([R_call], **kwargs):
            logger.error("Error installing R interface from the built package.")
            restore_makevars(makevars, old_makevars)
            return False
        logger.info("Installed R interface.")
    finally:
        restore_makevars(makevars, old_makevars)
    return True


if __name__ == "__main__":
    out = install()
    if out:
        logger.info("R interface installed.")
    else:
        raise Exception("Failed to install R interface.")



from yggdrasil.tests import assert_raises
from yggdrasil import components


def test_import_component():
    r"""Test dynamic import of component."""
    # Test use of default
    components.import_component('serializer')
    components.import_component('serializer')
    # Test explict type (but still new to registry)
    components.import_component('serializer', 'direct')
    components.import_component('serializer', 'direct')
    # Test using class name
    components.import_component('serializer', 'PandasSerialize')
    # Test access to file through comm (including error)
    components.import_component('comm', 'pickle')
    assert_raises(ValueError, components.import_component, 'comm', 'invalid')
    # Tests with registry suspended
    out = components.suspend_registry()
    components.import_component('serializer')
    components.import_component('serializer', 'direct')
    components.import_component('serializer', 'PandasSerialize')
    components.restore_registry(out)


def test_create_component():
    r"""Test dynamic creation of component instance."""
    components.create_component('serializer', seritype='direct')



import os
import copy
import pprint
import yaml
from collections import OrderedDict
from jsonschema.exceptions import ValidationError
from yggdrasil import metaschema, backwards


_schema_fname = os.path.abspath(os.path.join(
    os.path.dirname(__file__), '.ygg_schema.yml'))
_schema = None


class SchemaDict(OrderedDict):
    r"""OrderedDict subclass for ordering schemas on read in Python 2."""

    def __repr__(self):
        return pprint.pformat(dict(self))


def ordered_load(stream, object_pairs_hook=SchemaDict, **kwargs):
    r"""Load YAML document from a file using a specified class to represent
    mapping types that allows for ordering.

    Args:
        stream (file): File stream to load the schema YAML from.
        object_pairs_hook (type, optional): Class that should be used to
            represent loaded maps. Defaults to SchemaDict.
        **kwargs: Additional keyword arguments are passed to decode_yaml.

    Returns:
        object: Result of ordered load.

    """
    kwargs['sorted_dict_type'] = object_pairs_hook
    out = metaschema.encoder.decode_yaml(stream, **kwargs)
    out = backwards.as_str(out, recurse=True, allow_pass=True)
    return out


def ordered_dump(data, **kwargs):
    r"""Dump object as a YAML document, representing SchemaDict objects as
    mapping type.

    Args:
        data (object): Python object that should be dumped.
        **kwargs: Additional keyword arguments are passed to encode_yaml.

    Returns:
        str: YAML document representating data.

    """
    kwargs['sorted_dict_type'] = [SchemaDict, OrderedDict]
    return metaschema.encoder.encode_yaml(data, **kwargs)


def clear_schema():
    r"""Clear global schema."""
    global _schema
    _schema = None


def init_schema(fname=None):
    r"""Initialize global schema."""
    global _schema
    if _schema is None:
        _schema = load_schema(fname)


def create_schema():
    r"""Create a new schema from the registry."""
    from yggdrasil.components import init_registry
    try:
        old_env = os.environ.get('YGG_RUNNING_YGGSCHEMA', None)
        os.environ['YGG_RUNNING_YGGSCHEMA'] = '1'
        x = SchemaRegistry(init_registry())
    finally:
        if old_env is None:
            del os.environ['YGG_RUNNING_YGGSCHEMA']
        else:  # pragma: no cover
            os.environ['YGG_RUNNING_YGGSCHEMA'] = old_env
    return x


def load_schema(fname=None):
    r"""Return the yggdrasil schema for YAML options.

    Args:
        fname (str, optional): Full path to the file that the schema should be
            loaded from. If the file dosn't exist, it is created. Defaults to
            _schema_fname.

    Returns:
        dict: yggdrasil YAML options.

    """
    if fname is None:
        fname = _schema_fname
    if not os.path.isfile(fname):
        x = create_schema()
        x.save(fname)
    return SchemaRegistry.from_file(fname)


def get_schema(fname=None):
    r"""Return the yggdrasil schema for YAML options.

    Args:
        fname (str, optional): Full path to the file that the schema should be
            loaded from. If the file dosn't exist, it is created. Defaults to
            _schema_fname.

    Returns:
        dict: yggdrasil YAML options.

    """
    global _schema
    if fname is None:
        init_schema()
        out = _schema
    else:
        out = load_schema(fname)
    return out


class ComponentSchema(object):
    r"""Schema information for one component.

    Args:
        schema_type (str): The name of the component.
        subtype_key (str): The name of the schema property/class attribute
            that should be used to differentiate between subtypes of this
            component.
        schema_registry (SchemaRegistry, optional): Registry of schemas
            that this schema is dependent on.
        **kwargs: Additional keyword arguments are entries in the component
            schema.

    Args:
        schema_type (str): The name of the component.
        schema_registry (SchemaRegistry): Registry of schemas.
        subtype_key (str): Schema property that is used to differentiate between
            subtypes of this component.
        schema_subtypes (dict): Mapping between component class names and the
            associated values of the subtype_key property for this component.

    """
    # _subtype_keys = {'model': 'language', 'comm': 'commtype', 'file': 'filetype',
    #                  'connection': 'connection_type'}

    def __init__(self, schema_type, subtype_key,
                 schema_registry=None, schema_subtypes=None):
        self._storage = SchemaDict()
        self._base_schema = None
        self.schema_registry = schema_registry
        self.schema_type = schema_type
        self.subtype_key = subtype_key
        if schema_subtypes is None:
            schema_subtypes = {}
        self.schema_subtypes = schema_subtypes
        super(ComponentSchema, self).__init__()

    def get_subtype_schema(self, subtype, unique=False, relaxed=False):
        r"""Get the schema for the specified subtype.

        Args:
            subtype (str): Component subtype to return schema for. If 'base',
                the schema for evaluating the component base will be returned.
            unique (bool, optional): If True, the returned schema will only
                contain properties that are specific to the specified subtype.
                If subtype is 'base', these will be properties that are valid
                for all of the registerd subtypes. Defaults to False.

        Returns:
            dict: Schema for specified subtype.

        """
        if subtype == 'base':
            out = copy.deepcopy(self._base_schema)
            # Add additional properties that apply to specific subtypes
            if not unique:
                out['additionalProperties'] = False
                for x in self._storage.values():
                    for k, v in x['properties'].items():
                        if (k != self.subtype_key) and (k not in out['properties']):
                            out['properties'][k] = copy.deepcopy(v)
        else:
            if subtype not in self._storage:
                s2c = self.subtype2class
                if subtype in s2c:
                    subtype = s2c[subtype]
            out = copy.deepcopy(self._storage[subtype])
            # Remove properties that apply to all subtypes
            if unique:
                out['additionalProperties'] = True
                if 'required' in out:
                    out['required'] = list(set(out['required'])
                                           - set(self._base_schema['required']))
                    if not out['required']:
                        del out['required']
                for k in self._base_schema['properties'].keys():
                    if (k != self.subtype_key) and (k in out['properties']):
                        del out['properties'][k]
                if not out['properties']:  # pragma: no cover
                    del out['properties']
        if relaxed:
            out['additionalProperties'] = True
        return out

    def get_schema(self, relaxed=False):
        r"""Get the schema defining this component.

        Args:
            relaxed (bool, optional): If True, the returned schema (and any
                definitions it includes) are relaxed to allow for objects with
                objects with additional properties to pass validation. Defaults
                to False.

        Returns:
            dict: Schema for this component.

        """
        r"""dict: Schema for this component."""
        out = {'description': 'Schema for %s components.' % self.schema_type,
               'title': self.schema_type,
               'allOf': [self.get_subtype_schema('base', relaxed=relaxed),
                         {'anyOf': [self.get_subtype_schema(x, unique=True)
                                    for x in self._storage.keys()]}]}
        return out

    @property
    def schema(self):
        r"""dict: Schema for this component."""
        return self.get_schema()

    @property
    def full_schema(self):
        r"""dict: Schema for evaluating YAML input file that fully specifies
        the properties for each component."""
        # TODO: Could be simplified to just 'anyOf' for subtypes, but need
        # to reconcile that with schema normalization which uses the
        # position in the schema
        out = {'description': 'Schema for %s components.' % self.schema_type,
               'title': self.schema_type,
               'allOf': [self.get_subtype_schema('base', unique=True),
                         {'anyOf': [self.get_subtype_schema(x)
                                    for x in self._storage.keys()]}]}
        return out

    @classmethod
    def from_schema(cls, schema, schema_registry=None):
        r"""Construct a ComponentSchema from a schema.

        Args:
            schema (dict): Schema.

        Returns:
            ComponentSchema: Schema with information from schema.

        """
        schema_type = schema['title']
        subt_schema = schema['allOf'][1]['anyOf']
        # Determine subtype key
        subt_overlap = set(list(subt_schema[0]['properties'].keys()))
        subt_props = copy.deepcopy(subt_overlap)
        for v in subt_schema[1:]:
            ikeys = set(list(v['properties'].keys()))
            subt_props |= ikeys
            subt_overlap &= ikeys
        assert(len(subt_overlap) == 1)
        subtype_key = list(subt_overlap)[0]
        assert(subtype_key in schema['allOf'][0]['properties'])
        # Initialize schema
        out = cls(schema_type, subtype_key, schema_registry=schema_registry)
        out._base_schema = schema['allOf'][0]
        for v in subt_schema:
            out._storage[v['title']] = v
            subtypes = v['properties'][out.subtype_key]['enum']
            out.schema_subtypes[v['title']] = subtypes
        # Remove subtype specific properties
        for k in subt_props:
            if k != out.subtype_key:
                del out._base_schema['properties'][k]
        out._base_schema['additionalProperties'] = True
        # Update subtype properties with general properties
        for x in out._storage.values():
            for k, v in out._base_schema['properties'].items():
                if k != out.subtype_key:
                    x['properties'][k] = copy.deepcopy(v)
            x['additionalProperties'] = False
        return out

    @classmethod
    def from_registry(cls, schema_type, schema_classes, **kwargs):
        r"""Construct a ComponentSchema from a registry entry.

        Args:
            schema_type (str): Name of component type to build.
            schema_classes (list): List of classes for the component type.
            **kwargs: Additional keyword arguments are passed to the class
                __init__ method.

        Returns:
            ComponentSchema: Schema with information from classes.

        """
        out = None
        for x in schema_classes.values():
            if out is None:
                out = cls(schema_type, x._schema_subtype_key, **kwargs)
            out.append(x)
        return out

    @property
    def properties(self):
        r"""list: Valid properties for this component."""
        return sorted(list(self.get_subtype_schema('base')['properties'].keys()))

    def get_subtype_properties(self, subtype):
        r"""Get the valid properties for a specific subtype.

        Args:
            subtype (str): Name of the subtype to get keys for.

        Returns:
            list: Valid properties for the specified subtype.

        """
        return sorted(list(self.get_subtype_schema(subtype)['properties'].keys()))

    @property
    def class2subtype(self):
        r"""dict: Mapping from class to list of subtypes."""
        return self.schema_subtypes

    @property
    def subtype2class(self):
        r"""dict: Mapping from subtype to class."""
        out = {}
        for k, v in self.schema_subtypes.items():
            for iv in v:
                out[iv] = k
        return out

    @property
    def default_subtype(self):
        r"""str: Default subtype."""
        return self._base_schema['properties'][self.subtype_key].get(
            'default', None)

    @property
    def subtypes(self):
        r"""list: All subtypes for this schema type."""
        out = []
        for v in self.schema_subtypes.values():
            out += v
        return sorted(list(set(out)))

    @property
    def classes(self):
        r"""list: All available classes for this schema."""
        return sorted([k for k in self.schema_subtypes.keys()])

    def append(self, comp_cls):
        r"""Append component class to the schema.

        Args:
            comp_cls (class): Component class that should be added.

        """
        assert(comp_cls._schema_type == self.schema_type)
        assert(comp_cls._schema_subtype_key == self.subtype_key)
        name = comp_cls.__name__
        # Append subtype
        subtype_list = getattr(comp_cls, '_%s' % self.subtype_key, None)
        if not isinstance(subtype_list, list):
            subtype_list = [subtype_list]
        subtype_list += getattr(comp_cls, '_%s_aliases' % self.subtype_key, [])
        self.schema_subtypes[name] = subtype_list
        # Create new schema for subtype
        new_schema = {'title': name,
                      'description': ('Schema for %s component %s subtype.'
                                      % (self.schema_type, subtype_list)),
                      'type': 'object',
                      'required': copy.deepcopy(comp_cls._schema_required),
                      'properties': copy.deepcopy(comp_cls._schema_properties),
                      'additionalProperties': False}
        if not new_schema['required']:
            del new_schema['required']
        new_schema['properties'].setdefault(self.subtype_key, {})
        new_schema['properties'][self.subtype_key]['enum'] = subtype_list
        # Add legacy properties
        if self.schema_type in ['connection', 'comm', 'file', 'model']:
            legacy_properties = {'driver': {'type': 'string',
                                            'description': (
                                                '[DEPRECATED] Name of driver '
                                                'class that should be used.')},
                                 'args': {'type': 'string',
                                          'description': (
                                              '[DEPRECATED] Arguments that should '
                                              'be provided to the driver.')}}
            for k, v in legacy_properties.items():
                if k not in new_schema['properties']:
                    new_schema['properties'][k] = v
        # Create base schema
        is_base = False
        if self._base_schema is None:
            is_base = True
            self._base_schema = dict(
                copy.deepcopy(new_schema),
                title='%s_base' % self.schema_type,
                description=('Base schema for all subtypes of %s components.'
                             % self.schema_type),
                dependencies={'driver': ['args']},
                additionalProperties=True)
        # Add description of subtype to subtype property after base to
        # prevent overwriting description of the property rather than the
        # property value.
        if comp_cls._schema_subtype_description is not None:
            new_schema['properties'][self.subtype_key]['description'] = (
                comp_cls._schema_subtype_description)
        # Update base schema, checking for compatiblity
        if not is_base:
            if 'required' in self._base_schema:
                self._base_schema['required'] = list(
                    set(self._base_schema['required'])
                    & set(new_schema['required']))
                if not self._base_schema['required']:  # pragma: no cover
                    del self._base_schema['required']
            prop_overlap = list(
                set([self.subtype_key])  # Force subtype keys to be included
                | (set(self._base_schema['properties'].keys())
                   & set(new_schema['properties'].keys())))
            new_base_prop = {}
            for k in prop_overlap:
                old = copy.deepcopy(self._base_schema['properties'][k])
                new = copy.deepcopy(new_schema['properties'][k])
                # Don't compare descriptions or properties defining subtype
                if k != self.subtype_key:
                    old.pop('description', None)
                    new.pop('description', None)
                    if old != new:  # pragma: debug
                        raise ValueError(
                            ("Schema for property '%s' of class '%s' "
                             "is %s, which differs from the existing "
                             "base class value (%s). Check that "
                             "another class dosn't have a conflicting "
                             "definition of the same property.")
                            % (k, comp_cls, new, old))
                # Assign original copy that includes description
                new_base_prop[k] = self._base_schema['properties'][k]
                if k == self.subtype_key:
                    new_base_prop[k]['enum'] = sorted(list(
                        set(new_base_prop[k]['enum']) | set(new['enum'])))
            self._base_schema['properties'] = new_base_prop
        self._storage[name] = copy.deepcopy(new_schema)
        # Verify that the schema is valid
        metaschema.validate_schema(self.schema)


class SchemaRegistry(object):
    r"""Registry of schema's for different integration components.

    Args:
        registry (dict, optional): Dictionary of registered components.
            Defaults to None and the registry will be empty.
        required (list, optional): Components that are required. Defaults to
            ['comm', 'file', 'model', 'connection']. Ignored if registry is None.

    Raises:
        ValueError: If registry is provided and one of the required components
            is missing.

    """
    
    _normalizers = {}
    _default_required_components = ['comm', 'file', 'model', 'connection']

    def __init__(self, registry=None, required=None):
        super(SchemaRegistry, self).__init__()
        self._cache = {}
        self._storage = SchemaDict()
        if required is None:
            required = self._default_required_components
        self.required_components = required
        if registry is not None:
            for k in required:
                if k not in registry:
                    raise ValueError("Component %s required." % k)
            # Create schemas for each component
            for k, v in registry.items():
                icomp = ComponentSchema.from_registry(k, v, schema_registry=self)
                self.add(k, icomp)

    def add(self, k, v):
        r"""Add a new component schema to the registry."""
        self._cache = {}
        self._storage[k] = v
        metaschema.validate_schema(self.schema)

    def get(self, k, *args, **kwargs):
        r"""Return a component schema from the registry."""
        return self._storage.get(k, *args, **kwargs)

    def get_definitions(self, relaxed=False):
        r"""Get schema definitions for the registered components.

        Args:
            relaxed (bool, optional): If True, the returned schema (and any
                definitions it includes) are relaxed to allow for objects with
                objects with additional properties to pass validation. Defaults
                to False.

        Returns:
            dict: Schema defintiions for each of the registered components.

        """
        if relaxed:
            cache_key = 'relaxed_definitions'
        else:
            cache_key = 'definitions'
        if cache_key not in self._cache:
            out = {k: v.get_schema(relaxed=relaxed)
                   for k, v in self._storage.items()}
            for k in self.required_components:
                out.setdefault(k, {'type': 'string'})
            self._cache[cache_key] = out
        return copy.deepcopy(self._cache[cache_key])

    @property
    def definitions(self):
        r"""dict: Schema definitions for different components."""
        return self.get_definitions()

    @property
    def schema(self):
        r"""dict: Schema for evaluating YAML input file."""
        if 'schema' not in self._cache:
            out = {'title': 'YAML Schema',
                   'description': 'Schema for yggdrasil YAML input files.',
                   'type': 'object',
                   'definitions': self.definitions,
                   'required': ['models'],
                   'additionalProperties': False,
                   'properties': SchemaDict(
                       [('models', {'type': 'array',
                                    'items': {'$ref': '#/definitions/model'},
                                    'minItems': 1}),
                        ('connections',
                         {'type': 'array',
                          'items': {'$ref': '#/definitions/connection'}})])}
            self._cache['schema'] = out
        return copy.deepcopy(self._cache['schema'])

    @property
    def full_schema(self):
        r"""dict: Schema for evaluating YAML input file that fully specifies
        the properties for each component."""
        if 'full_schema' not in self._cache:
            out = self.schema
            for k, v in self._storage.items():
                out['definitions'][k] = v.full_schema
            self._cache['full_schema'] = out
        return copy.deepcopy(self._cache['full_schema'])

    def __getitem__(self, k):
        return self.get(k)

    def keys(self):
        return self._storage.keys()

    def __eq__(self, other):
        if not hasattr(other, 'schema'):
            return False
        return (self.schema == other.schema)

    @classmethod
    def from_file(cls, fname):
        r"""Create a SchemaRegistry from a file.

        Args:
            fname (str): Full path to the file the schema should be loaded from.

        """
        out = cls()
        out.load(fname)
        return out

    def load(self, fname):
        r"""Load schema from a file.

        Args:
            fname (str): Full path to the file the schema should be loaded from.

        """
        with open(fname, 'r') as f:
            contents = f.read()
            schema = ordered_load(contents, Loader=yaml.SafeLoader)
        if schema is None:
            raise Exception("Failed to load schema from %s" % fname)
        # Create components
        for k, v in schema.get('definitions', {}).items():
            icomp = ComponentSchema.from_schema(v, schema_registry=self)
            self.add(k, icomp)

    def save(self, fname):
        r"""Save the schema to a file.

        Args:
            fname (str): Full path to the file the schema should be saved to.
            schema (dict): yggdrasil YAML options.

        """
        out = self.schema
        with open(fname, 'w') as f:
            ordered_dump(out, stream=f, Dumper=yaml.SafeDumper)

    def validate(self, obj, **kwargs):
        r"""Validate an object against this schema.

        Args:
            obj (object): Object to valdiate.
            **kwargs: Additional keyword arguments are passed to validate_instance.

        """
        if kwargs.get('normalize', False):
            kwargs.setdefault('normalizers', self._normalizers)
            # kwargs.setdefault('no_defaults', True)
            kwargs.setdefault('schema_registry', self)
        return metaschema.validate_instance(obj, self.schema, **kwargs)

    def validate_component(self, comp_name, obj):
        r"""Validate an object against a specific component.

        Args:
            comp_name (str): Name of the component to validate against.
            obj (object): Object to validate.

        """
        comp_schema = self.get_component_schema(comp_name)
        return metaschema.validate_instance(obj, comp_schema)

    def normalize(self, obj, backwards_compat=False, **kwargs):
        r"""Normalize an object against this schema.

        Args:
            obj (object): Object to normalize.
            **kwargs: Additional keyword arguments are passed to normalize_instance.

        Returns:
            object: Normalized object.

        """
        kwargs.setdefault('normalizers', self._normalizers)
        kwargs.setdefault('required_defaults', True)
        kwargs.setdefault('no_defaults', True)
        kwargs.setdefault('schema_registry', self)
        return metaschema.normalize_instance(obj, self.full_schema, **kwargs)

    # def is_valid(self, obj):
    #     r"""Determine if an object is valid under this schema.

    #     Args:
    #         obj (object): Object to valdiate.

    #     Returns:
    #         bool: True if the object is valid, False otherwise.

    #     """
    #     try:
    #         self.validate(obj)
    #     except ValidationError:
    #         return False
    #     return True

    def is_valid_component(self, comp_name, obj):
        r"""Determine if an object is a valid represenation of a component.

        Args:
            comp_name (str): Name of the component to validate against.
            obj (object): Object to validate.

        Returns:
            bool: True if the object is valid, False otherwise.

        """
        try:
            self.validate_component(comp_name, obj)
        except ValidationError:
            return False
        return True

    def get_component_schema(self, comp_name, subtype=None, relaxed=False):
        r"""Get the schema for a certain component.

        Args:
            comp_name (str): Name of the component to get the schema for.
            subtype (str, optional): Component subtype to get schema for.
                Defaults to None and the schema for evaluating any subtype of
                the specified component is returned.
            relaxed (bool, optional): If True, the returned schema (and any
                definitions it includes) are relaxed to allow for objects with
                objects with additional properties to pass validation. Defaults
                to False.

        Returns:
            dict: Schema for the specified component.

        """
        if comp_name not in self._storage:  # pragma: debug
            raise ValueError("Unrecognized component: %s" % comp_name)
        if subtype is None:
            out = self._storage[comp_name].get_schema(relaxed=relaxed)
        else:
            out = self._storage[comp_name].get_subtype_schema(subtype,
                                                              relaxed=relaxed)
        out['definitions'] = self.get_definitions(relaxed=relaxed)
        return out

    def get_component_keys(self, comp_name):
        r"""Get the properties associated with a certain component.

        Args:
            comp_name (str): Name of the component to return keys for.

        Returns:
            list: All of the valid properties for the specified component.

        """
        return self._storage[comp_name].properties

    @classmethod
    def register_normalizer(cls, path):
        r"""Register a normalizer that will be applied to elements in the
        instance at the specified path.

        Args:
            path (tuple): Location in schema where normalizer will be applied.

        Returns:
            function: Decorator for registering the normalizer function.

        """
        if not isinstance(path, list):
            path_list = [path]
        else:
            path_list = path

        def _register_normalizer(func):
            for p in path_list:
                if p not in cls._normalizers:
                    cls._normalizers[p] = []
                cls._normalizers[p].append(func)
            return func

        return _register_normalizer


# The following are function add to allow backwards compatability of older
# yaml schemas
def rwmeth2filetype(rw_meth):
    r"""Get the alternate properties that corresponding to the old
    read_meth/write_meth keywords.

    Args:
        rw_meth (str): Read/write method name.

    Returns:
        dict: Property values equivalent to provided read/write method.

    """
    out = {}
    if rw_meth == 'all':
        out['filetype'] = 'binary'
    elif rw_meth == 'line':
        out['filetype'] = 'ascii'
    elif rw_meth == 'table_array':
        out['filetype'] = 'table'
        out['as_array'] = True
    else:
        out['filetype'] = rw_meth
    return out


def cdriver2commtype(driver):
    r"""Convert a connection driver to a file type.

    Args:
        driver (str): The name of the connection driver.

    Returns:
        str: The corresponding file type for the driver.

    """
    _legacy = {'InputDriver': 'default',
               'OutputDriver': 'default',
               'ZMQInputDriver': 'zmq',
               'ZMQOutputDriver': 'zmq',
               'IPCInputDriver': 'ipc',
               'IPCOutputDriver': 'ipc',
               'RMQInputDriver': 'rmq',
               'RMQOutputDriver': 'rmq',
               'RMQAsyncInputDriver': 'rmq_async',
               'RMQAsyncOutputDriver': 'rmq_async'}
    if driver in _legacy:
        return _legacy[driver]
    raise ValueError("Unknown driver: '%s'" % driver)

    
def cdriver2filetype(driver):
    r"""Convert a connection driver to a file type.

    Args:
        driver (str): The name of the connection driver.

    Returns:
        str: The corresponding file type for the driver.

    """
    _legacy = {'FileInputDriver': 'binary',
               'FileOutputDriver': 'binary',
               'AsciiMapInputDriver': 'map',
               'AsciiMapOutputDriver': 'map',
               'AsciiFileInputDriver': 'ascii',
               'AsciiFileOutputDriver': 'ascii',
               'AsciiTableInputDriver': 'table',
               'AsciiTableOutputDriver': 'table',
               'PandasFileInputDriver': 'pandas',
               'PandasFileOutputDriver': 'pandas',
               'PickleFileInputDriver': 'pickle',
               'PickleFileOutputDriver': 'pickle',
               'PlyFileInputDriver': 'ply',
               'PlyFileOutputDriver': 'ply',
               'ObjFileInputDriver': 'obj',
               'ObjFileOutputDriver': 'obj',
               'MatInputDriver': 'mat',
               'MatOutputDriver': 'mat'}
    if driver in _legacy:
        return _legacy[driver]
    raise ValueError("%s is not a registered connection driver." % driver)


def migrate_keys(from_dict, to_dict, exclude_key_list=None, include_key_list=None):
    r"""Migrate keys from one component to another that are not in a list
    of predefined keys.

    Args:
         from_dict (dict): Component dictionary to migrate keys from.
         to_dict (list): List of component dictionaries to migrate keys to. If
             this is an empty list, keys will not be migrated.
         exclude_key_list (list, optional): List of keys in from_dict that
             should not be migrated to to_dict. All keys in include_key_list
             that are not in this list are moved. Defaults to None and no keys
             are excluded.
         include_key_list (list, optional): List of keys that should be migrated
             from from_dict to to_dict dictionaries. If not provided, all keys
             in from_dict that are not in exclude_key_list are moved. Defaults
             to None and all keys in from_dict are included.

    """
    assert(isinstance(to_dict, list))
    if len(to_dict) == 0:
        return
    if exclude_key_list is None:
        exclude_key_list = []
    if include_key_list is None:
        include_key_list = list(from_dict.keys())
    for k in include_key_list:
        if (k not in from_dict) or (k in exclude_key_list):
            continue
        v = from_dict.pop(k)
        for d in to_dict:
            d.setdefault(k, v)


def standardize(instance, keys, is_singular=False, suffixes=None, altkeys=None):
    r"""Standardize a component such that each key contains a list of dictionaries.

    Args:
        instance (dict): Component to standardize.
        keys (list): Keys to standardize in the instance.
        is_singular (bool, optional): If False, the keys are assumed to be plural
            and singular alternatives are also checked. If True, the keys are
            assumed to be singular and plural alternatives are also checked.
            Defaults to False.
        suffixes (list, optional): Suffixes to add to the keys to get a set of
            alternate keys that should also be checked. Defaults to None as is
            ignored.
        altkeys (list, optional): List of lists of alternate keys. Defaults to None.

    """
    for k in keys:
        if k not in instance:
            instance[k] = []
        if not isinstance(instance[k], list):
            instance[k] = [instance[k]]
    # Get list of alternate keys from suffixes and plurality
    if altkeys is None:
        altkeys = []
    if suffixes is not None:
        for s in suffixes:
            altkeys.append(['%s%s' % (k, s) for k in keys])
            if is_singular:
                altkeys.append(['%ss%s' % (k, s) for k in keys])
                altkeys.append(['%s%ss' % (k, s) for k in keys])
            else:
                altkeys.append(['%s%s' % (k[:-1], s) for k in keys])
                altkeys.append(['%s%ss' % (k[:-1], s) for k in keys])
    if is_singular:
        altkeys.append(['%ss' % k for k in keys])
    else:
        altkeys.append([k[:-1] for k in keys])
    # Add components listed under alternate keys
    for ialtkeys in altkeys:
        for k, kalt in zip(keys, ialtkeys):
            if kalt in instance:
                if isinstance(instance[kalt], list):
                    instance[k] += instance.pop(kalt)
                else:
                    instance[k].append(instance.pop(kalt))
    # Handle strings
    for k in keys:
        for i in range(len(instance[k])):
            if isinstance(instance[k][i], str):
                instance[k][i] = {'name': instance[k][i]}


@SchemaRegistry.register_normalizer(tuple())
def _normalize_root(normalizer, value, instance, schema):
    r"""Decorate normalizer."""
    # if getattr(normalizer, 'schema_registry', None) is None:
    #     normalizer.schema_registry = get_schema()
    if getattr(normalizer, 'iodict', None) is None:
        normalizer.iodict = {'inputs': {}, 'outputs': {}, 'connections': [],
                             'input_drivers': [], 'output_drivers': [], 'pairs': [],
                             'inputs_extra': {}, 'outputs_extra': {}}
    standardize(instance, ['models', 'connections'])
    return instance


@SchemaRegistry.register_normalizer(('models', 0))
def _normalize_modelio_first(normalizer, value, instance, schema):
    r"""Normalizes set of model inputs/outputs before each input/output is normalized."""
    if isinstance(instance, dict):
        standardize(instance, ['inputs', 'outputs'])
        for io in ['inputs', 'outputs']:
            for x in instance[io]:
                x.setdefault('working_dir', instance['working_dir'])
    return instance


@SchemaRegistry.register_normalizer([('models', 0, 'inputs', 0),
                                     ('models', 0, 'outputs', 0)])
def _normalize_modelio_elements(normalizer, value, instance, schema):
    r"""Normalize case of models singular."""
    io = normalizer.current_schema_path[2]
    # Register io if dict set
    iodict = getattr(normalizer, 'iodict', None)
    s = getattr(normalizer, 'schema_registry', None)
    if (iodict is not None) and isinstance(instance, dict) and ('name' in instance):
        # Register io if dict set
        if instance['name'] not in iodict[io]:
            iodict[io][instance['name']] = instance
            # Move non-comm keywords to a buffer
            if (s is not None):
                comm_keys = s.get_component_keys('comm')
                type_keys = list(metaschema.get_metaschema()['properties'].keys())
                extra_keys = {}
                migrate_keys(instance, [extra_keys], comm_keys + type_keys)
                iodict['%s_extra' % io][instance['name']] = extra_keys
                # type_dict = {}
                # migrate_keys(instance, [type_dict], comm_keys)
                # instance.setdefault('datatype', {})
                # instance['datatype'].update(type_dict)
        # Add driver to list
        if ('driver' in instance) and ('args' in instance):
            opp_map = {'inputs': 'output', 'outputs': 'input'}
            for i, (opp_arg, opp_name) in enumerate(iodict['%s_drivers' % opp_map[io]]):
                if instance['args'] == opp_arg:
                    if io == 'inputs':
                        iodict['pairs'].append(
                            (iodict['%s_drivers' % opp_map[io]].pop(i)[1],
                             instance['name']))
                    else:  # pragma: debug
                        # This won't be called because inputs are processed first
                        # but this code is here for symmetries sake
                        iodict['pairs'].append(
                            (instance['name'],
                             iodict['%s_drivers' % opp_map[io]].pop(i)[1]))
                    break
            else:
                iodict['%s_drivers' % io[:-1]].append(
                    (instance['args'], instance['name']))
    return instance


@SchemaRegistry.register_normalizer(('connections',))
def _normalize_connio_base(normalizer, value, instance, schema):
    r"""Normalizes list of connections, adding those represented by
    multiple drivers."""
    if normalizer.current_path != normalizer.current_schema_path:
        return instance
    # Build connections from input/output drivers
    iodict = getattr(normalizer, 'iodict', None)
    if (iodict is not None):
        new_connections = []
        # Create direct connections from output to input
        for (oname, iname) in iodict['pairs']:
            oyml = iodict['outputs'][oname]
            iyml = iodict['inputs'][iname]
            conn = dict(input=oname, output=iname)
            new_connections.append(([oyml, iyml], conn))
            oyml['commtype'] = cdriver2commtype(oyml['driver'])
            iyml['commtype'] = cdriver2commtype(iyml['driver'])
            oyml.pop('working_dir', None)
            iyml.pop('working_dir', None)
        # File input
        for k, v in iodict['input_drivers']:
            iyml = iodict['inputs'][v]
            fyml = dict(name=k, filetype=cdriver2filetype(iyml['driver']))
            if iyml.get('as_array', False):
                # TODO: This should not be an exception
                fyml['as_array'] = True
            conn = dict(input=fyml, output=v)
            new_connections.append(([iyml], conn))
        # File output
        for k, v in iodict['output_drivers']:
            oyml = iodict['outputs'][v]
            fyml = dict(name=k, filetype=cdriver2filetype(oyml['driver']))
            if oyml.get('as_array', False):
                # TODO: This should not be an exception
                fyml['as_array'] = True
            conn = dict(output=fyml, input=v)
            new_connections.append(([oyml], conn))
        # Transfer keyword arguments from input/output to connection
        for ymls, conn in new_connections:
            for y in ymls:
                del y['driver'], y['args']
            iodict['connections'].append(conn)
            instance.append(conn)
        # Empty registry of orphan input/output drivers
        for k in ['input_drivers', 'output_drivers', 'pairs']:
            iodict[k] = []
    return instance


@SchemaRegistry.register_normalizer(('connections', 0))
def _normalize_connio_first(normalizer, value, instance, schema):
    r"""Normalizes set of connection before each connection is normalized."""
    if isinstance(instance, dict):
        standardize(instance, ['inputs', 'outputs'], suffixes=['_file', '_files'],
                    altkeys=[['from', 'to']])
        # Handle indexed inputs/outputs
        for io in ['inputs', 'outputs']:
            pruned = []
            pruned_names = []
            for x in instance[io]:
                if (':' in x['name']) and (not os.path.isabs(x['name'])):
                    name = x['name'].split(':')[0]
                    x['name'] = name
                if x['name'] not in pruned_names:
                    pruned_names.append(x['name'])
                    pruned.append(x)
            instance[io] = pruned
        # Move non-comm properties from model inputs/outputs
        s = getattr(normalizer, 'schema_registry', None)
        iodict = getattr(normalizer, 'iodict', None)
        if (s is not None) and (iodict is not None):
            opp_map = {'inputs': 'outputs', 'outputs': 'inputs'}
            comm_keys = s.get_component_keys('comm')
            conn_keys = s.get_component_keys('connection')
            target_files = []
            for io in ['inputs', 'outputs']:
                for x in instance[io]:
                    y = iodict['%s_extra' % opp_map[io]].get(x['name'], None)
                    if y is None:
                        target_files.append(x)
                        continue
                    y_keys = list(y.keys())
                    for k in y_keys:
                        val = y.pop(k)
                        if k == 'translator':
                            instance.setdefault(k, [])
                            if not isinstance(val, (list, tuple)):
                                val = [val]
                            instance[k] += val
                        else:
                            instance.setdefault(k, val)
            # Move everything but comm keywords down to files, then move
            # comm keywords down to connection inputs and outputs.
            migrate_keys(instance, target_files, conn_keys + comm_keys)
            instance.pop('working_dir', None)
            migrate_keys(instance, instance['inputs'] + instance['outputs'], conn_keys)
    return instance


@SchemaRegistry.register_normalizer([('connections', 0, 'inputs', 0, 0),
                                     ('connections', 0, 'outputs', 0, 0)])
def _normalize_connio_elements_comm(normalizer, value, instance, schema):
    r"""Normalize connection inputs/outputs as comms."""
    io = normalizer.current_schema_path[2]
    if isinstance(instance, dict):
        # Check to see if is file
        iodict = getattr(normalizer, 'iodict', None)
        opp_map = {'inputs': 'outputs', 'outputs': 'inputs'}
        if iodict is not None:
            if (instance['name'] in iodict[opp_map[io]]):
                opp_comm = iodict[opp_map[io]][instance['name']]
                s = getattr(normalizer, 'schema_registry', None)
                if s is not None:
                    comm_keys = s.get_component_keys('comm')
                    for k in comm_keys:
                        if k in opp_comm:
                            instance.setdefault(k, opp_comm[k])
    return instance


@SchemaRegistry.register_normalizer([('connections', 0, 'inputs', 1, 0),
                                     ('connections', 0, 'outputs', 1, 0)])
def _normalize_connio_elements_file(normalizer, value, instance, schema):
    r"""Normalize connection inputs/outputs as files."""
    io = normalizer.current_schema_path[2]
    if isinstance(instance, dict):
        # Check to see if is file
        iodict = getattr(normalizer, 'iodict', None)
        opp_map = {'inputs': 'outputs', 'outputs': 'inputs'}
        if iodict is not None:
            if (((instance['name'] not in iodict[opp_map[io]])
                 and ('filetype' not in instance))):
                instance['filetype'] = schema['properties']['filetype']['default']
    return instance


@SchemaRegistry.register_normalizer(('connections', 1))
def _normalize_connio_last(normalizer, value, instance, schema):
    r"""Normalize set of connections after they have been normalized."""
    if isinstance(instance, dict):
        # Check that files properly specified
        s = getattr(normalizer, 'schema_registry', None)
        if s is not None:
            is_file = {}
            for io in ['inputs', 'outputs']:
                all = [s.is_valid_component('file', x) for x in instance[io]]
                is_file[io] = (sum(all) == len(all))
            if is_file['inputs'] and is_file['outputs']:
                raise RuntimeError(("Both the input and output for this connection "
                                    + "appear to be files:\n%s"
                                    % pprint.pformat(instance)))
        # Copy file keys from partner comm(s) to the file comm(s)
        comm_keys = s.get_component_keys('comm')
        opp_map = {'inputs': 'outputs', 'outputs': 'inputs'}
        for io in ['inputs', 'outputs']:
            if is_file[io]:
                for x in instance[opp_map[io]]:
                    migrate_keys(x, instance[io], comm_keys)
    return instance


@SchemaRegistry.register_normalizer(('models', 0))
def _normalize_model_driver(normalizer, value, instance, schema):
    r"""Normalizes older style of specifying driver rather than language."""
    if isinstance(instance, dict):
        s = getattr(normalizer, 'schema_registry', None)
        if s is not None:
            if ('language' not in instance) and ('driver' in instance):
                if instance['driver'] == 'GCCModelDriver':
                    # TODO: Fix this properly, checking the extension to
                    # distinguish between C and C++
                    if isinstance(instance['args'], list):
                        args_ext = os.path.splitext(instance['args'][0])[-1]
                    else:
                        args_ext = os.path.splitext(instance['args'])[-1]
                    from yggdrasil.drivers.CPPModelDriver import CPPModelDriver
                    if args_ext in CPPModelDriver.language_ext:
                        instance['driver'] = 'CPPModelDriver'
                    else:
                        instance['driver'] = 'CModelDriver'
                class2language = s['model'].class2subtype
                instance['language'] = class2language[instance.pop('driver')][0]
    return instance


@SchemaRegistry.register_normalizer([('connections', 0, 'inputs', 1, 0),
                                     ('connections', 0, 'outputs', 1, 0)])
def _normalize_rwmeth(normalizer, value, instance, schema):
    r"""Normalize older style of specifying 'read_meth' or 'write_meth' instead
    of filetype."""
    if isinstance(instance, dict):
        # Replace old read/write methd with filetype
        for k in ['read_meth', 'write_meth']:
            val = instance.pop(k, None)
            if (((val is not None)
                 and (instance.get('filetype', None) in [None, 'binary']))):
                instance.update(rwmeth2filetype(val))
    return instance


@SchemaRegistry.register_normalizer([('connections', 0, 'inputs', 1, 0),
                                     ('connections', 0, 'outputs', 1, 0)])
def _normalize_ascii_table(normalizer, value, instance, schema):
    r"""Normalize the older style arguments for ascii table connections."""
    if isinstance(instance, dict):
        alias_keys = [('column_names', 'field_names'),
                      ('column_units', 'field_units'),
                      ('column', 'delimiter')]
        for old, new in alias_keys:
            if old in instance:
                instance.setdefault(new, instance.pop(old))
    return instance


@SchemaRegistry.register_normalizer([('connections', 0, 'inputs', 1, 0),
                                     ('connections', 0, 'outputs', 1, 0)])
def _normalize_serializer(normalizer, value, instance, schema):
    r"""Normalize the serializer if the information is in the file."""
    if ((isinstance(instance, dict) and ('serializer' not in instance)
         and (instance.get('filetype', None) in [None, 'binary']))):
        s = getattr(normalizer, 'schema_registry', None)
        if s is not None:
            comm_keys = s.get_component_keys('comm')
            seri_keys = s.get_component_keys('serializer')
            serializer = {}
            migrate_keys(instance, [serializer], include_key_list=seri_keys,
                         exclude_key_list=comm_keys)
            if serializer:
                instance['serializer'] = serializer
    return instance


@SchemaRegistry.register_normalizer([('models', 0, 'inputs', 0),
                                     ('models', 0, 'outputs', 0),
                                     ('connections', 0, 'inputs', 0, 0),
                                     ('connections', 0, 'outputs', 0, 0)])
def _normalize_datatype(normalizer, value, instance, schema):
    r"""Normalize the datatype if the type information is in the comm."""
    if isinstance(instance, dict) and ('datatype' not in instance):
        type_keys = list(metaschema.get_metaschema()['properties'].keys())
        # Don't include args in type_keys if driver in the instance
        if ('driver' in instance) and ('args' in type_keys):
            type_keys.remove('args')
        datatype = {}
        migrate_keys(instance, [datatype], include_key_list=type_keys)
        if datatype:
            instance['datatype'] = datatype
    return instance
